from __future__ import annotations

import uuid
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

from DAJIN2.core import preprocess
from DAJIN2.utils import config, fastx_handler, io


def parse_arguments(arguments: dict) -> tuple:
    genome_urls = defaultdict(str)
    if arguments.get("genome"):
        genome_urls.update(
            {"genome": arguments["genome"], "blat": arguments["blat"], "goldenpath": arguments["goldenpath"]}
        )

    return (
        Path(arguments["sample"]),
        Path(arguments["control"]),
        Path(arguments["allele"]),
        arguments["name"],
        arguments["threads"],
        genome_urls,
        uuid.uuid4().hex,
    )


def convert_input_paths_to_posix(sample: str, control: str, allele: str) -> tuple:
    sample = io.convert_to_posix(sample)
    control = io.convert_to_posix(control)
    allele = io.convert_to_posix(allele)

    return sample, control, allele


def create_temporal_directory(name: str, control_name: str) -> Path:
    tempdir = Path(config.TEMP_ROOT_DIR, name)
    Path(tempdir, "cache", ".igvjs", control_name).mkdir(parents=True, exist_ok=True)

    return tempdir


def check_caches(tempdir: Path, path_allele: str, genome_url: str) -> bool:
    is_cache_hash = preprocess.exists_cached_hash(tempdir=tempdir, path=path_allele)
    is_cache_genome = preprocess.exists_cached_genome(tempdir=tempdir, genome=genome_url)

    return is_cache_hash and is_cache_genome


def get_genome_coordinates(genome_urls: dict, fasta_alleles: dict, is_cache_genome: bool, tempdir: Path) -> dict:
    genome_coordinates = {
        "genome": genome_urls["genome"],
        "chrom_size": 0,
        "chrom": "control",
        "start": 0,
        "end": len(fasta_alleles["control"]) - 1,
        "strand": "+",
    }
    if genome_urls["genome"]:
        if is_cache_genome:
            genome_coordinates = next(io.read_jsonl(Path(tempdir, "cache", "genome_coordinates.jsonl")))
        else:
            genome_coordinates = preprocess.fetch_coordinates(
                genome_coordinates, genome_urls, fasta_alleles["control"]
            )
            genome_coordinates["chrom_size"] = preprocess.fetch_chromosome_size(genome_coordinates, genome_urls)
            io.write_jsonl([genome_coordinates], Path(tempdir, "cache", "genome_coordinates.jsonl"))

    return genome_coordinates


@dataclass(frozen=True)
class FormattedInputs:
    path_sample: Path
    path_control: Path
    path_allele: Path
    sample_name: str
    control_name: str
    fasta_alleles: dict[str, str]
    tempdir: Path
    genome_coordinates: dict[str, str]
    threads: int
    uuid: str


def format_inputs(arguments: dict) -> FormattedInputs:
    path_sample, path_control, path_allele, name, threads, genome_urls, uuid = parse_arguments(arguments)
    path_sample, path_control, path_allele = convert_input_paths_to_posix(path_sample, path_control, path_allele)
    sample_name = fastx_handler.extract_filename(path_sample)
    control_name = fastx_handler.extract_filename(path_control)
    fasta_alleles = fastx_handler.dictionize_allele(path_allele)
    tempdir = create_temporal_directory(name, control_name)
    is_cache_genome = check_caches(tempdir, path_allele, genome_urls["genome"])
    genome_coordinates = get_genome_coordinates(genome_urls, fasta_alleles, is_cache_genome, tempdir)

    return FormattedInputs(
        path_sample,
        path_control,
        path_allele,
        io.sanitize_name(sample_name),
        io.sanitize_name(control_name),
        fasta_alleles,
        tempdir,
        genome_coordinates,
        threads,
        uuid,
    )



from __future__ import annotations

import cstag


def get_index_of_insertions(ref_cstag: str) -> set:
    index_of_insertions = set()
    start = 0
    for cs in cstag.split(ref_cstag):
        if cs.startswith("+"):
            index_of_insertions |= set(range(start, start + len(cs[1:])))
            start += len(cs[1:])
        elif cs.startswith("="):
            start += len(cs[1:])
        elif cs.startswith("*"):
            start += 1
        elif cs.startswith("-"):
            continue
    return index_of_insertions


def split_cstag(cs_tag: str) -> list[str]:
    cstag_split = []
    for cs in cstag.split(cs_tag):
        if cs.startswith("=") or cs.startswith("-") or cs.startswith("+"):
            op = cs[0]
            cstag_split += [op + c for c in list(cs[1:])]
        elif cs.startswith("*"):
            cstag_split.append(cs)
    return cstag_split


def apply_insertion(cs_split: list[str], index_of_insertions: set) -> list[str]:
    idx = -1
    cs_insertion = []
    for cs in cs_split:
        if not cs.startswith("+"):
            idx += 1
        if idx in index_of_insertions:
            if cs.startswith("-"):
                continue
            cs_insertion.append("+" + cs[-1].lower())
        else:
            cs_insertion.append(cs)
    return cs_insertion


def convert_to_cstag(cs_insertion: list[str]) -> str:
    cs_tag = []
    prev_op = "@"
    for cs in cs_insertion:
        if cs.startswith(prev_op) and prev_op != "*":
            cs_tag.append(cs[1:])
            continue
        prev_op = cs[0]
        cs_tag.append(cs)
    return "".join(cs_tag)


def reflect_ref_insertion_to_query(ref_cstag: str, que_cstag: str) -> str:
    """
    Translate the cstag of the query corresponding to the insertion site of the reference into an insertion.

    Example:
        ref_cstag = "=AC+aaa=GT"
        que_cstag = "=ACAAAGT"
        reflect_ref_insertion_to_query(ref_cstag, que_cstag)  # -> "=AC+aaa=GT"
    """
    index_of_insertions = get_index_of_insertions(ref_cstag)
    cs_split = split_cstag(que_cstag)
    cs_insertion = apply_insertion(cs_split, index_of_insertions)
    return convert_to_cstag(cs_insertion)



from __future__ import annotations

import os

os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"


import argparse
import logging
import shutil
import sys
from copy import deepcopy
from itertools import groupby
from pathlib import Path

from DAJIN2 import gui, view
from DAJIN2.core import core
from DAJIN2.utils import config, input_validator, io, multiprocess, report_generator


def generate_report(name: str, logger: logging.Logger) -> None:
    report_generator.report(name)
    logger.info(f"\N{PARTY POPPER} Finished! Open {config.DAJIN_RESULTS_DIR}/{name} to see the report.")


################################################################################
# Single mode
################################################################################


def execute_single_mode(arguments: dict[str]):
    # Set logging to export log to stderr and file
    path_logfile = config.get_logfile()
    logger = config.set_logging(path_logfile)
    logger.info(f"\N{RUNNER} Start running DAJIN2 version {config.DAJIN_VERSION}")
    logger.info(f"\N{PERSONAL COMPUTER} {' '.join(sys.argv)}")

    # Validate input files
    input_validator.validate_files(arguments["sample"], arguments["control"], arguments["allele"])
    if arguments.get("genome"):
        arguments.update(input_validator.validate_genome_and_fetch_urls(arguments["genome"]))

    # Run DAJIN2
    core.execute_control(arguments)
    core.execute_sample(arguments)

    # Finish call
    generate_report(arguments["name"], logger)
    shutil.move(path_logfile, Path("DAJIN_Results", arguments["name"]))

    # Remove temporary files
    if not arguments["debug"]:
        shutil.rmtree(Path("DAJIN_Results", ".tempdir", arguments["name"]))


################################################################################
# Batch mode
################################################################################


def validate_headers_of_batch_file(headers: set[str], filepath: str) -> None:
    """Validate the headers of a batch file."""
    required_headers = {"sample", "control", "allele", "name"}
    accepted_headers = {"sample", "control", "allele", "name", "genome"}

    if not required_headers.issubset(headers):
        raise ValueError(f'{filepath} must contain "sample", "control", "allele" and "name" in the header')

    if not headers.issubset(accepted_headers):
        raise ValueError(
            f'Accepted header names of {filepath} are "sample", "control", "allele", "name", or "genome".'
        )


def create_argument_dict(args: dict, cache_urls_genome: dict, is_control: bool) -> dict[str, str]:
    """Create a dictionary of arguments from the given headers and group."""
    args_update = deepcopy(args)

    args_update["threads"] = 1  # Set the number of threads to 1 for batch mode

    # Assign the "sample" field depending on whether it's a control or not
    if is_control:
        args_update["sample"] = args_update["control"]
    else:
        if args_update["sample"] == args_update["control"]:
            return {}  # Return an empty dict to indicate a skipped group

    if args_update.get("genome"):
        args_update.update(cache_urls_genome[args_update["genome"]])

    return args_update


def run_DAJIN2(
    groups: list[dict[str, str]], cache_urls_genome: dict, is_control: bool = True, num_workers: int = 1
) -> None:
    contents = []
    for args in groups:
        args = create_argument_dict(args, cache_urls_genome, is_control)
        if args:  # Add args to contents only if it's not an empty dict
            contents.append(args)

    # Return a list of unique dictionaries
    contents_unique = [dict(item) for item in {frozenset(d.items()) for d in contents}]

    contents_unique.sort(key=lambda x: x["sample"])

    if is_control:
        multiprocess.run(core.execute_control, contents_unique, num_workers)
    else:
        multiprocess.run(core.execute_sample, contents_unique, num_workers)


def execute_batch_mode(arguments: dict[str]):
    path_batchfile = io.convert_to_posix(arguments["file"])

    if not Path(path_batchfile).exists():
        raise FileNotFoundError(f"'{path_batchfile}' does not exist.")

    records = io.load_batchfile(path_batchfile)

    # Validate Column of the batch file
    headers = set(records[0].keys())
    validate_headers_of_batch_file(headers, path_batchfile)

    # Validate contents and fetch genome urls
    cache_urls_genome = {}
    records.sort(key=lambda x: x["name"])
    for _, groups in groupby(records, key=lambda x: x["name"]):
        for args in groups:
            # Validate contents in the batch file
            input_validator.validate_files(args["sample"], args["control"], args["allele"])
            # Validate genome and fetch urls
            if args.get("genome") and args["genome"] not in cache_urls_genome:
                urls_genome = input_validator.validate_genome_and_fetch_urls(args["genome"])
                cache_urls_genome[args["genome"]] = urls_genome

    # Run DAJIN2
    for name, groups in groupby(records, key=lambda x: x["name"]):
        groups: list[dict[str, str]] = list(groups)
        # Set logging to export log to stderr and file
        config.reset_logging()
        path_logfile = config.get_logfile()
        logger = config.set_logging(path_logfile)
        logger.info(f"\N{RUNNER} Start running DAJIN2 version {config.DAJIN_VERSION}")
        logger.info(f"\N{PERSONAL COMPUTER} {' '.join(sys.argv)}")

        # Run DAJIN2
        run_DAJIN2(groups, cache_urls_genome, is_control=True, num_workers=arguments["threads"])
        run_DAJIN2(groups, cache_urls_genome, is_control=False, num_workers=arguments["threads"])

        # Finish call
        generate_report(name, logger)
        shutil.move(path_logfile, Path("DAJIN_Results", name))
        if not arguments["debug"]:
            shutil.rmtree(Path("DAJIN_Results", ".tempdir", name))


def execute():
    parser = argparse.ArgumentParser()

    ###############################################################################
    # Single mode
    ###############################################################################

    parser.add_argument("-s", "--sample", type=str, help="Path to a sample directory including FASTQ file")
    parser.add_argument("-c", "--control", type=str, help="Path to a control directory including FASTQ file")
    parser.add_argument("-a", "--allele", type=str, help="Path to a FASTA file")
    parser.add_argument("-n", "--name", type=str, help="Output directory name", default="Results")
    parser.add_argument(
        "-g", "--genome", type=str, default="", help="Reference genome ID (e.g hg38, mm39) [default: '']"
    )
    parser.add_argument("-t", "--threads", type=int, default=1, help="Number of threads [default: 1]")
    parser.add_argument("-v", "--version", action="version", version=f"DAJIN2 version {config.DAJIN_VERSION}")
    parser.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)

    ###############################################################################
    # Batch mode
    ###############################################################################

    def batchmode(args):
        arguments = {}
        arguments["file"] = args.file
        arguments["threads"] = input_validator.update_threads(int(args.threads))
        arguments["debug"] = args.debug
        execute_batch_mode(arguments)

    subparser = parser.add_subparsers()
    parser_batch = subparser.add_parser("batch", help="DAIJN2 batch mode")
    parser_batch.add_argument("-f", "--file", required=True, type=str, help="CSV or Excel file.")
    parser_batch.add_argument("-t", "--threads", default=1, type=int, help="Number of threads [default: 1]")
    parser_batch.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)
    parser_batch.set_defaults(handler=batchmode)

    ###############################################################################
    # GUI mode
    ###############################################################################

    def guimode(args):
        gui.execute()

    parser_gui = subparser.add_parser("gui", help="DAIJN2 GUI mode")
    parser_gui.set_defaults(handler=guimode)

    ###############################################################################
    # View mode
    ###############################################################################

    def viewmode(args):
        view.execute(args.name)

    parser_view = subparser.add_parser("view", help="DAIJN2 View mode to launch igvjs")
    parser_view.add_argument("-n", "--name", required=True, type=str, help="Output name of the report")
    parser_view.set_defaults(handler=viewmode)

    ###############################################################################
    # Parse arguments
    ###############################################################################

    args = parser.parse_args()

    if hasattr(args, "handler"):
        args.handler(args)
    else:
        if args.sample is None:
            raise ValueError("the following arguments are required: -s/--sample")
        if args.control is None:
            raise ValueError("the following arguments are required: -c/--control")
        if args.allele is None:
            raise ValueError("the following arguments are required: -a/--allele")
        if args.name is None:
            raise ValueError("the following arguments are required: -n/--name")
        arguments = {}
        arguments["sample"] = args.sample
        arguments["control"] = args.control
        arguments["allele"] = args.allele
        arguments["name"] = args.name
        arguments["genome"] = args.genome
        arguments["threads"] = input_validator.update_threads(int(args.threads))
        arguments["debug"] = args.debug

        execute_single_mode(arguments)


if __name__ == "__main__":
    execute()



from __future__ import annotations

import bisect
import re
from collections import defaultdict
from pathlib import Path
from typing import Generator

from DAJIN2.utils import config

"""
To suppress the following warnings from `scipy.wilcoxon`:
UserWarning: Exact p-value calculation does not work if there are zeros.
"""
config.set_warnings_ignore()


import numpy as np
from sklearn.neural_network import MLPClassifier

from DAJIN2.core.preprocess.homopolymer_handler import extract_sequence_errors_in_homopolymer_loci
from DAJIN2.utils import io


def count_indels(midsv_sample: Generator[dict], sequence: str) -> dict[str, list[int]]:
    len_sequence = len(sequence)
    count = {"=": [0] * len_sequence, "+": [0] * len_sequence, "-": [0] * len_sequence, "*": [0] * len_sequence}
    for samp in midsv_sample:
        for i, cs in enumerate(samp["CSSPLIT"].split(",")):
            if cs == "N" or re.search(r"a|c|g|t|n", cs):
                continue
            if cs.startswith("="):
                count["="][i] += 1
            elif cs.startswith("+"):
                count["+"][i] += 1
            elif cs.startswith("-"):
                count["-"][i] += 1
            elif cs.startswith("*"):
                count["*"][i] += 1
    return count


def normalize_indels(count: dict[str, list[int]]) -> dict[str, np.array]:
    count_normalized = {}
    match_count = np.array(count["="])
    for mut, indel_count in count.items():
        numerator = np.array(indel_count)
        denominator = numerator + match_count
        count_normalized[mut] = np.where(denominator == 0, 0, numerator / denominator * 100)
    return count_normalized


def minimize_mutation_counts(
    indels_control: dict[str, np.array], indels_sample: dict[str, np.array]
) -> dict[str, np.array]:
    """
    In cases where control has a larger value than sample, adjust the value of sample to match that of control.
    """
    indels_control_minimized = {}
    for mut in {"+", "-", "*"}:
        indels_control_minimized[mut] = np.minimum(indels_control[mut], indels_sample[mut])
    return indels_control_minimized


def split_kmer(indels: dict[str, np.array], kmer: int = 11) -> dict[str, np.array]:
    results = defaultdict(list)
    center = kmer // 2
    for mut, value in indels.items():
        for i in range(len(value)):
            if center <= i <= len(value) - center:
                start = i - center
                if kmer % 2 == 0:
                    end = i + center
                else:
                    end = i + center + 1
                results[mut].append(value[start:end])
            else:
                results[mut].append(np.array([0] * kmer))
    return results


###########################################################
# Extract Anomalous Loci
###########################################################


def cosine_distance(x: list[float], y: list[float]) -> float:
    # Add 1e-6 to avoid division by zero when calculating cosine similarity
    x = np.array(x) + 1e-6
    y = np.array(y) + 1e-6
    return 1 - float(np.dot(x, y) / (np.linalg.norm(x) * np.linalg.norm(y)))


def is_dissimilar_loci(values_sample, values_control, index: int, is_consensus: bool = False) -> bool:
    # If 'sample' has more than 20% variation compared to 'control' in consensus mode, unconditionally set it to 'dissimilar loci'. This is set to counteract cases where, when evaluating cosine similarity during significant deletions, values exceedingly close to 1 can occur even if not observed in the control (e.g., control = [1,1,1,1,1], sample = [100,100,100,100,100] -> cosine similarity = 1).
    if values_sample[index] - values_control[index] > 20:
        if not is_consensus or values_sample[index] > 50:
            return True
        else:
            return False

    # Subset 10 bases around index.
    x = values_sample[index : index + 10]
    y = values_control[index : index + 10]

    x_slice = x[1:]
    y_slice = y[1:]

    distance = cosine_distance(x, y)
    distance_slice = cosine_distance(x_slice, y_slice)

    return distance > 0.05 and distance / (distance + distance_slice) > 0.9


def detect_anomalies(values_sample, values_control, threshold: float, is_consensus: bool = False) -> set[int]:
    """
    Detect anomalies and return indices of outliers.
    """

    rng = np.random.default_rng(seed=1)

    random_size = 10_000
    control_size = len(values_control)
    total_size = random_size + control_size

    randoms = rng.uniform(0, 100, random_size)
    randoms_error = np.clip(randoms + rng.uniform(0, threshold, random_size), 0, 100)
    randoms_mutation = np.clip(randoms + rng.uniform(threshold, 100, random_size), 0, 100)

    values_error = np.clip(values_control + rng.uniform(0, threshold, control_size), 0, 100)
    values_mutation = np.clip(values_control + rng.uniform(threshold, 100, control_size), 0, 100)

    matrix_error_randoms = np.array([randoms, randoms_error]).T
    matrix_error_control = np.array([values_control, values_error]).T
    matrix_error = np.concatenate([matrix_error_randoms, matrix_error_control], axis=0)

    matrix_mutation_randoms = np.array([randoms, randoms_mutation]).T
    matrix_mutation_control = np.array([values_control, values_mutation]).T
    matrix_mutation = np.concatenate([matrix_mutation_randoms, matrix_mutation_control], axis=0)

    X = np.concatenate([matrix_error, matrix_mutation], axis=0)
    y = [0] * (total_size) + [1] * (total_size)

    clf = MLPClassifier(solver="lbfgs", alpha=1e-5, hidden_layer_sizes=(5, 2), random_state=1)
    clf.fit(X, y)

    results = clf.predict(np.array([values_control, values_sample]).T)

    candidate_loci = {i for i, v in enumerate(results) if v == 1}

    return {i for i in candidate_loci if is_dissimilar_loci(values_sample, values_control, i, is_consensus)}


def extract_anomal_loci(
    indels_normalized_sample,
    indels_normalized_control,
    thresholds: dict[str, float],
    is_consensus: bool = False,
) -> dict[str, set[int]]:
    """Extract outlier loci compareing indel counts between sample and control."""
    anomal_loci = {}
    for mut in {"+", "-", "*"}:
        values_sample = indels_normalized_sample[mut]
        values_control = indels_normalized_control[mut]
        idx_outliers = detect_anomalies(values_sample, values_control, thresholds[mut], is_consensus)
        anomal_loci[mut] = idx_outliers
    return anomal_loci


###########################################################
# Homolopolymer region
###########################################################


def discard_errors_in_homopolymer(loci: dict[str, set[int]], errors: dict[str, set[int]]) -> dict[str, set[int]]:
    """Remove detected errors in homopolymer regions from the candidate loci."""
    return {mut: loci[mut] - errors[mut] for mut in {"+", "-", "*"}}


###########################################################
# Merge contiguous insertions/deletions
###########################################################


def count_elements_within_range(arr, lower_bound, upper_bound):
    """
    Counts the number of elements within a specified range in a sorted array.
    """
    start_index = bisect.bisect_left(arr, lower_bound)
    end_index = bisect.bisect_right(arr, upper_bound)
    return end_index - start_index


def merge_index_of_consecutive_indel(mutation_loci: dict[str, set[int]]) -> dict[str, set[int]]:
    """Treat as contiguous indels if there are insertions/deletions within five bases of each other"""
    mutation_loci_merged = {}

    # Reflect point mutations as they are
    mutation_loci_merged["*"] = mutation_loci["*"]

    # Merge if indels are within 10 bases
    for mut in ["+", "-"]:
        idx_indel = sorted(mutation_loci[mut])
        idx_indel_merged = set(idx_indel)
        for i in range(len(idx_indel) - 1):
            idx_1 = idx_indel[i]
            idx_2 = idx_indel[i + 1]
            # If everything from idx_1 to idx_2 is already considered as indels, then skip it.
            if count_elements_within_range(idx_indel, idx_1 + 1, idx_2 - 1) == idx_2 - idx_1 + 1:
                continue
            if idx_1 + 10 > idx_2:
                for i in range(idx_1 + 1, idx_2):
                    idx_indel_merged.add(i)
        mutation_loci_merged[mut] = idx_indel_merged

    # Additional logic for mutation enrichment within 10 bases on both ends
    for mut in ["+", "-"]:
        idx_indel = sorted(mutation_loci_merged[mut])
        idx_indel_merged = set(idx_indel)
        for i in range(len(idx_indel) - 1):
            idx_1 = idx_indel[i]
            idx_2 = idx_indel[i + 1]
            # If everything from idx_1 to idx_2 is already considered as indels, then skip it.
            if count_elements_within_range(idx_indel, idx_1 + 1, idx_2 - 1) == idx_2 - idx_1 + 1:
                continue
            # If the distance between idx_1 and idx_2 is more than 20 bases, then skip it.
            if idx_1 + 20 < idx_2:
                continue
            count_left = count_elements_within_range(idx_indel, idx_1 - 11, idx_1 - 1)
            count_right = count_elements_within_range(idx_indel, idx_2 + 1, idx_2 + 11)

            # If 8 out of the 10 bases at both ends are indels,
            # then everything from idx_1 to idx_2 will be considered as indels.

            if count_left >= 8 and count_right >= 8:
                for i in range(idx_1 + 1, idx_2):
                    idx_indel_merged.add(i)
        mutation_loci_merged[mut] = idx_indel_merged

    return mutation_loci_merged


###########################################################
# formatters
###########################################################


def summarize_indels(path_midsv: Path, sequence: str) -> tuple:
    """Returns indels, coverages, normalized indels, and kmer indels."""
    indels_count = count_indels(io.read_jsonl(path_midsv), sequence)
    indels_normalized = normalize_indels(indels_count)

    return indels_count, indels_normalized


def merge_loci(dissimilar_loci: dict[str, set], anomal_loci: dict[str, set]) -> dict[str, set]:
    mutation_loci = {}
    for mut in {"+", "-", "*"}:
        mutation_loci[mut] = dissimilar_loci[mut] | anomal_loci[mut]
    return mutation_loci


def add_knockin_loci(candidate_loci: dict[str, set], knockin_loci: set):
    mutation_loci = {}
    for mut in {"+", "-", "*"}:
        mutation_loci[mut] = candidate_loci[mut] | knockin_loci
    return mutation_loci


def transpose_mutation_loci(mutation_loci: dict[str, set[int]], sequence: str) -> list[set[str]]:
    len_sequence = len(sequence)
    mutation_loci_transposed = [set() for _ in range(len_sequence)]
    for mut, idx_mutation in mutation_loci.items():
        for i, loci in enumerate(mutation_loci_transposed):
            if i in idx_mutation:
                loci.add(mut)
    return mutation_loci_transposed


###########################################################
# main
###########################################################


def cache_indels_count(ARGS, is_control: bool = False) -> None:
    dirname = ARGS.control_name if is_control else ARGS.sample_name
    for allele, sequence in ARGS.fasta_alleles.items():
        path_mutation_loci = Path(ARGS.tempdir, dirname, "mutation_loci", allele)
        path_mutation_loci.mkdir(parents=True, exist_ok=True)

        if not is_control:
            prefix = ARGS.sample_name
        else:
            path_insertion = Path(ARGS.tempdir, ARGS.control_name, "midsv", allele, f"{ARGS.sample_name}.jsonl")
            if path_insertion.exists():
                prefix = ARGS.sample_name
            else:
                prefix = ARGS.control_name

        if Path(path_mutation_loci, f"{prefix}_count.pickle").exists():
            continue

        path_midsv = Path(ARGS.tempdir, dirname, "midsv", allele, f"{prefix}.jsonl")
        indels_count, indels_normalized = summarize_indels(path_midsv, sequence)
        io.save_pickle(indels_count, Path(path_mutation_loci, f"{prefix}_count.pickle"))
        io.save_pickle(indels_normalized, Path(path_mutation_loci, f"{prefix}_normalized.pickle"))


def extract_mutation_loci(
    sequence: str,
    path_indels_normalized_sample: Path,
    path_indels_normalized_control: Path,
    path_knockin: Path,
    thresholds: dict[str, float] = None,
    is_consensus: bool = False,
) -> list[set[str]]:
    if thresholds is None:
        thresholds = {"*": 0.5, "-": 0.5, "+": 0.5}
    indels_normalized_sample = io.load_pickle(path_indels_normalized_sample)

    # Extract candidate mutation loci
    indels_normalized_control = minimize_mutation_counts(
        io.load_pickle(path_indels_normalized_control), indels_normalized_sample
    )
    anomal_loci: dict[str, set[int]] = extract_anomal_loci(
        indels_normalized_sample, indels_normalized_control, thresholds, is_consensus
    )

    # Extract error loci in homopolymer regions
    errors_in_homopolymer = extract_sequence_errors_in_homopolymer_loci(
        sequence, indels_normalized_sample, indels_normalized_control, anomal_loci
    )
    mutation_loci = discard_errors_in_homopolymer(anomal_loci, errors_in_homopolymer)

    # Merge all mutations and knockin loci
    if path_knockin.exists():
        knockin_loci = io.load_pickle(path_knockin)
        mutation_loci = add_knockin_loci(mutation_loci, knockin_loci)

    mutation_loci_merged = merge_index_of_consecutive_indel(mutation_loci)
    mutation_loci_transposed = transpose_mutation_loci(mutation_loci_merged, sequence)
    return mutation_loci_transposed


def cache_mutation_loci(ARGS, is_control: bool = False) -> None:
    cache_indels_count(ARGS, is_control)

    if is_control:
        return None

    for allele, sequence in ARGS.fasta_alleles.items():
        path_mutation_sample = Path(ARGS.tempdir, ARGS.sample_name, "mutation_loci", allele)
        path_mutation_control = Path(ARGS.tempdir, ARGS.control_name, "mutation_loci", allele)

        path_output_mutation_loci = Path(path_mutation_sample, "mutation_loci.pickle")
        if path_output_mutation_loci.exists():
            continue

        file_name = f"{ARGS.sample_name}_normalized.pickle"
        if not Path(path_mutation_control, file_name).exists():
            file_name = f"{ARGS.control_name}_normalized.pickle"

        path_indels_normalized_control = Path(path_mutation_control, file_name)

        path_indels_normalized_sample = Path(path_mutation_sample, f"{ARGS.sample_name}_normalized.pickle")
        path_knockin = Path(ARGS.tempdir, ARGS.sample_name, "knockin_loci", allele, "knockin.pickle")

        mutation_loci: list[set[str]] = extract_mutation_loci(
            sequence, path_indels_normalized_sample, path_indels_normalized_control, path_knockin
        )

        io.save_pickle(mutation_loci, path_output_mutation_loci)



from __future__ import annotations

from itertools import groupby
from pathlib import Path

import numpy as np
from sklearn.cluster import MiniBatchKMeans

from DAJIN2.core.consensus.similarity_searcher import cache_selected_control_by_similarity
from DAJIN2.core.preprocess.mutation_extractor import extract_mutation_loci, minimize_mutation_counts, summarize_indels
from DAJIN2.utils import io

"""
Most of the code reuses `preprocess.cache_mutation_loci`.
This code differs from `preprocess.cache_mutation_loci` in that it dynamically extracts control sequences similar to alleles after clustering, and sets thresholds for distinguishing between sequence errors and mutations.
- Extracts reads similar to alleles after clustering to minimize the impact of anomalies in the control sequences.
- Since alleles after clustering essentially assume a single allele, the threshold setting in preprocess is not 10% but is set to a higher value.
"""


def get_thresholds(path_indels_normalized_sample, path_indels_normalized_control) -> dict[str, float]:
    """
    For the consensus threshold, the minimum threshold for "sample - control" was set to be at least 10%.
    """
    indels_normalized_sample = io.load_pickle(path_indels_normalized_sample)
    indels_normalized_control = io.load_pickle(path_indels_normalized_control)
    indels_normalized_minimize_control = minimize_mutation_counts(indels_normalized_control, indels_normalized_sample)
    thresholds = {}
    for mut in {"+", "-", "*"}:
        values_sample = indels_normalized_sample[mut]
        values_control = indels_normalized_minimize_control[mut]
        values_subtract = values_sample - values_control
        kmeans = MiniBatchKMeans(n_clusters=2, random_state=0).fit(values_subtract.reshape(-1, 1))
        threshold = kmeans.cluster_centers_.mean()
        thresholds[mut] = max(threshold, 10)
    return thresholds


def extract_path_control(tempdir: Path, control_name: str, sample_name: str, allele: str) -> Path:
    # Define potential file paths
    paths = [
        Path(tempdir, control_name, "midsv", allele, f"{control_name}.jsonl"),
        Path(tempdir, control_name, "midsv", allele, f"{sample_name}.jsonl"),
    ]
    # Return the first path that exists
    for path in paths:
        if path.exists():
            return path


def extract_path_n_filtered_control(
    tempdir: Path, control_name: str, sample_name: str, path_control: Path, allele: str
) -> Path:
    """
    Filter out reads with N enriched in the control.
    """
    path_output = Path(tempdir, control_name, "consensus", allele, f"{sample_name}_n_filtered.jsonl")
    if path_output.exists():
        return path_output

    path_output.parent.mkdir(parents=True, exist_ok=True)

    midsv_control = io.read_jsonl(path_control)
    n_counts = np.array([sum(1 if cs == "N" else 0 for cs in c["CSSPLIT"].split(",")) for c in midsv_control])

    kmeans = MiniBatchKMeans(n_clusters=2, random_state=0).fit(n_counts.reshape(-1, 1))
    threshold = kmeans.cluster_centers_.mean()
    labels = np.where(n_counts <= threshold, True, False)
    midsv_filtered_control = (m for m, label in zip(io.read_jsonl(path_control), labels) if label)
    io.write_jsonl(midsv_filtered_control, path_output)

    return path_output


def cache_normalized_indels(ARGS, path_consensus_sample: Path, allele: str) -> None:
    sequence = ARGS.fasta_alleles[allele]

    path_control = extract_path_control(ARGS.tempdir, ARGS.control_name, ARGS.sample_name, allele)
    path_control_filtered = extract_path_n_filtered_control(
        ARGS.tempdir, ARGS.control_name, ARGS.sample_name, path_control, allele
    )

    cache_selected_control_by_similarity(ARGS, path_control_filtered, path_consensus_sample, allele)

    path_midsv_similar_control = Path(path_consensus_sample.parent, "control.jsonl")

    _, indels_normalized_sample = summarize_indels(path_consensus_sample, sequence)
    _, indels_normalized_control = summarize_indels(path_midsv_similar_control, sequence)

    io.save_pickle(indels_normalized_sample, Path(path_consensus_sample.parent, "normalized_sample.pickle"))
    io.save_pickle(indels_normalized_control, Path(path_consensus_sample.parent, "normalized_control.pickle"))


def cache_mutation_loci(ARGS, clust_sample: list[dict]) -> None:
    # Separate clusters by label and cache them
    clust_sample.sort(key=lambda x: [x["ALLELE"], x["LABEL"]])
    for (allele, label), group in groupby(clust_sample, key=lambda x: [x["ALLELE"], x["LABEL"]]):
        path_consensus = Path(ARGS.tempdir, ARGS.sample_name, "consensus", allele, str(label))
        path_consensus.mkdir(parents=True, exist_ok=True)

        path_consensus_sample = Path(path_consensus, f"{ARGS.sample_name}_sample.jsonl")
        io.write_jsonl(group, path_consensus_sample)

        cache_normalized_indels(ARGS, path_consensus_sample, allele)

        # Extract and cache mutation loci
        path_indels_normalized_sample = Path(path_consensus, "normalized_sample.pickle")
        path_indels_normalized_control = Path(path_consensus, "normalized_control.pickle")
        sequence = ARGS.fasta_alleles[allele]
        path_knockin = Path(ARGS.tempdir, ARGS.sample_name, "knockin_loci", allele, "knockin.pickle")

        thresholds = get_thresholds(path_indels_normalized_sample, path_indels_normalized_control)

        is_consensus = True

        mutation_loci = extract_mutation_loci(
            sequence,
            path_indels_normalized_sample,
            path_indels_normalized_control,
            path_knockin,
            thresholds,
            is_consensus,
        )

        io.save_pickle(mutation_loci, Path(path_consensus, "mutation_loci.pickle"))



from __future__ import annotations

from collections import defaultdict
from itertools import groupby
from pathlib import Path

import pysam

from DAJIN2.utils import io, sam_handler


def recalculate_sam_coodinates_to_reference(sam: list[list[str]], GENOME_COODINATES: dict) -> list[str]:
    """Recalculate SAM genomic coordinates with the reference genome, not with the FASTA_ALLELE"""
    sam_headers = [s for s in sam if s[0].startswith("@")]
    sam_contents = [s for s in sam if not s[0].startswith("@")]
    for s in sam_headers:
        if s[0] != "@SQ":
            continue
        s[1] = f'SN:{GENOME_COODINATES["chrom"]}'
        s[2] = f'LN:{GENOME_COODINATES["chrom_size"]}'
    for s in sam_contents:
        s[2] = GENOME_COODINATES["chrom"]
    if GENOME_COODINATES["strand"] == "-":
        sam_contents = sam_handler.revcomp_sam(sam_contents, GENOME_COODINATES["end"])
    else:
        for s in sam_contents:
            s[3] = str(int(s[3]) + GENOME_COODINATES["start"] - 1)
    return sam_headers + sam_contents


def convert_pos_to_one_indexed(sam_lines: list[list[str]]) -> list[list[str]]:
    """Convert SAM POS from 0-indexed to 1-indexed"""

    def convert_line(line: list[str]) -> list[str]:
        if not line[0].startswith("@") and line[3] == "0":
            line[3] = "1"
        return line

    return [convert_line(line) for line in sam_lines]


def group_by_name(sam_contents: list[str], clust_sample: list[dict]) -> dict[list]:
    """Group alignments in map-ont.sam by allele name (NAME)"""
    sam_contents.sort()
    clust_sample_sorted = sorted(clust_sample, key=lambda x: x["QNAME"])

    qnames: set[str] = {c["QNAME"] for c in clust_sample_sorted}

    sam_groups = defaultdict(list)
    idx_sam_contents = 0
    idx_clust_sample = 0
    while idx_sam_contents < len(sam_contents) and idx_clust_sample < len(clust_sample_sorted):
        alignments_sam = sam_contents[idx_sam_contents][:-1]  # Discard CS tags to reduce file size
        alignments_clsut_sample = clust_sample_sorted[idx_clust_sample]
        qname_sam = alignments_sam[0]

        if qname_sam not in qnames:
            idx_sam_contents += 1
            continue

        if qname_sam == alignments_clsut_sample["QNAME"]:
            key = alignments_clsut_sample["NAME"]
            sam_groups[key].append(alignments_sam)
            idx_sam_contents += 1
        else:
            idx_clust_sample += 1

    return dict(sam_groups)


###############################################################################
# igvjs
###############################################################################


def subset_qnames(RESULT_SAMPLE, readnum: int = 100) -> dict[set[str]]:
    qnames_by_name = defaultdict(set)
    for name, group in groupby(RESULT_SAMPLE, key=lambda x: x["NAME"]):
        group = list(group)
        qnames = [res["QNAME"] for res in group[:readnum]]
        qnames_by_name[name] = set(qnames)
    return dict(qnames_by_name)


def subset_reads(sam_content: list[str], qnames: set[str]) -> list[str]:
    return [sam for sam in sam_content if sam[0] in qnames]


###############################################################################
# Output
###############################################################################


def write_sam_to_bam(sam: list[list[str]], path_sam: str | Path, path_bam: str | Path, threads: int = 1) -> None:
    formatted_sam = "\n".join("\t".join(s) for s in sam)
    Path(path_sam).write_text(formatted_sam + "\n")
    pysam.sort("-@", f"{threads}", "-o", str(path_bam), str(path_sam))
    pysam.index("-@", f"{threads}", str(path_bam))


def update_sam(sam: list, GENOME_COODINATES: dict = None) -> list:
    if GENOME_COODINATES is None:
        GENOME_COODINATES = {}

    sam_records = sam.copy()
    sam_records = sam_handler.remove_microhomology(sam_records)
    if GENOME_COODINATES["genome"]:
        return recalculate_sam_coodinates_to_reference(sam_records, GENOME_COODINATES)
    else:
        return convert_pos_to_one_indexed(sam_records)


def export_to_bam(TEMPDIR, NAME, GENOME_COODINATES, THREADS, UUID, RESULT_SAMPLE=None, is_control=False) -> None:
    path_sam_input = Path(TEMPDIR, NAME, "sam", "control", "map-ont.sam")
    if not path_sam_input.exists():  # In the case of short-read.
        path_sam_input = Path(TEMPDIR, NAME, "sam", "control", "sr.sam")

    sam_records = list(io.read_sam(path_sam_input))

    # Update sam
    sam_updated = update_sam(sam_records, GENOME_COODINATES)

    # Output SAM and BAM
    path_sam_output = Path(TEMPDIR, "report", "BAM", f"temp_{UUID}_{NAME}_control.sam")
    path_bam_output = Path(TEMPDIR, "report", "BAM", NAME, f"{NAME}.bam")
    write_sam_to_bam(sam_updated, path_sam_output, path_bam_output, THREADS)

    # Prepare SAM headers and contents
    sam_headers = [s for s in sam_updated if s[0].startswith("@")]
    sam_contents = [s for s in sam_updated if not s[0].startswith("@")]
    if is_control:
        qnames_100reads: set[str] = set(list({s[0] for s in sam_contents[:10000]})[:100])  # subset 100 reads
        sam_subset = [s for s in sam_updated if s[0] in qnames_100reads]
        path_sam_output = Path(TEMPDIR, "report", "BAM", f"temp_{UUID}_{NAME}_control_cache.sam")
        path_bam_output = Path(TEMPDIR, "cache", ".igvjs", NAME, "control.bam")
        write_sam_to_bam(sam_headers + sam_subset, path_sam_output, path_bam_output, THREADS)
    else:
        sam_groups = group_by_name(sam_contents, RESULT_SAMPLE)
        qnames_by_name = subset_qnames(RESULT_SAMPLE)
        # Output SAM and BAM
        for name, sam_content in sam_groups.items():
            # BAM
            path_sam_output = Path(TEMPDIR, "report", "BAM", f"temp_{UUID}_{name}.sam")
            path_bam_output = Path(TEMPDIR, "report", "BAM", NAME, f"{NAME}_{name}.bam")
            write_sam_to_bam(sam_headers + sam_content, path_sam_output, path_bam_output, THREADS)
            # igvjs
            sam_subset = subset_reads(sam_content, qnames_by_name[name])
            path_sam_output = Path(TEMPDIR, "report", "BAM", f"temp_{UUID}_{name}_subset.sam")
            path_bam_output = Path(TEMPDIR, "report", ".igvjs", NAME, f"{name}.bam")
            write_sam_to_bam(sam_headers + sam_subset, path_sam_output, path_bam_output, THREADS)

    # Remove temporary files
    sam_temp = Path(TEMPDIR, "report", "BAM").glob(f"temp_{UUID}*.sam")
    [s.unlink() for s in sam_temp]



from __future__ import annotations

import csv
import hashlib
import json
import pickle
import re
from io import BufferedReader
from pathlib import Path
from typing import Generator

import wslPath
from openpyxl import Workbook, load_workbook

###########################################################
# Input/Output
###########################################################


def read_sam(path_of_sam: str | Path) -> Generator[list]:
    with open(path_of_sam) as f:
        for line in f:
            yield line.strip().split("\t")


def load_pickle(file_path: Path):
    with open(file_path, "rb") as f:
        return pickle.load(f)


def save_pickle(data: object, file_path: Path) -> None:
    with open(file_path, "wb") as f:
        pickle.dump(data, f)


def read_jsonl(file_path: str | Path) -> Generator[dict]:
    with open(file_path) as f:
        for line in f:
            yield json.loads(line)


def write_jsonl(data: list[dict], file_path: str | Path) -> None:
    with open(file_path, "w") as f:
        for d in data:
            json_str = json.dumps(d)
            f.write(json_str + "\n")


def write_xlsx(data: list[dict[str, str]], file_path: str | Path) -> None:
    # Create a workbook and a worksheet
    wb = Workbook()
    ws = wb.active

    if not data:
        wb.save(file_path)
        return

    # Initialize headers with the keys of the first dictionary, maintaining order
    headers = list(data[0].keys())

    # Check for new keys in the subsequent dictionaries and append them to the headers list
    for item in data[1:]:
        for key in item.keys():
            if key not in headers:
                headers.append(key)

    # Write the headers to the first row
    ws.append(headers)

    # Write the data to Excel
    for item in data:
        row = [item.get(header, "") for header in headers]
        ws.append(row)

    # Save the file
    wb.save(file_path)


###########################################################
# Load batch file
###########################################################


def determine_file_type(file_path: str) -> str | None:
    """Determine if the file is an Excel or CSV file. Raise error for other types."""
    file_extension = Path(file_path).suffix
    if file_extension in [".xlsx", ".xls"]:
        return "excel"
    elif file_extension == ".csv":
        return "csv"
    else:
        raise ValueError("The provided file must be either an Excel or CSV file.")


def read_xlsx(file_path: str | Path) -> list[dict[str, str]]:
    """Load data from an Excel file."""
    wb = load_workbook(filename=file_path)
    ws = wb.active

    headers = list(next(ws.iter_rows(min_row=1, max_row=1, values_only=True)))

    records = []
    for row in ws.iter_rows(min_row=2, values_only=True):
        if all(element is None for element in row):  # Skip rows with all None values
            continue
        row_data = {headers[i]: (row[i] if i < len(row) else None) for i in range(len(headers))}
        records.append(row_data)

    return records


def read_csv(file_path: str | Path) -> list[dict[str, str]]:
    """Load data from a CSV file."""
    with open(file_path) as csvfile:
        header = [field.strip() for field in next(csv.reader(csvfile))]

        records = []
        for row in csv.reader(csvfile):
            if not row:  # Skip empty rows
                continue
            if all(element is None for element in row):  # Skip rows with all None values
                continue
            row_trimmed = [field.strip() for field in row]
            row_data = dict(zip(header, row_trimmed))
            records.append(row_data)

        return records


def load_batchfile(batchfile_path: str) -> list[dict[str, str]]:
    """Load data from either an Excel or CSV file."""
    file_type = determine_file_type(batchfile_path)
    if file_type == "excel":
        return read_xlsx(batchfile_path)
    elif file_type == "csv":
        return read_csv(batchfile_path)


###########################################################
# File hander
###########################################################


def count_newlines(filepath: str | Path) -> int:
    def read_in_chunks(file: BufferedReader, chunk_size: int = 2**16) -> Generator[bytes, None, None]:
        """Get a generator that reads a file in chunks and yields each chunk."""
        while True:
            chunk = file.read(chunk_size)
            if not chunk:
                break
            yield chunk

    # Open the file in binary mode and count the newline characters
    with open(filepath, "rb") as file:
        count = sum(chunk.count(b"\n") for chunk in read_in_chunks(file))

    return count


def convert_to_posix(path: str) -> str:
    if wslPath.is_windows_path(path):
        path = wslPath.to_posix(path)
    return path


def sanitize_name(file_name: Path | str) -> str:
    """
    Sanitize the file name by replacing invalid characters on Windows OS with '-'
    """
    file_name = str(file_name).strip()
    if not file_name:
        raise ValueError("Provided name is empty or consists only of whitespace")
    forbidden_chars = r'[<>:"/\\|?*\x00-\x1F .]'

    return re.sub(forbidden_chars, "-", file_name)


###########################################################
# Cache control hash
###########################################################


def cache_control_hash(tempdir: Path, path_allele: str | Path) -> None:
    """
    Generate a hash value from the content of the provided file and cache it.

    Parameters:
    - tempdir: Path object, the temporary directory path.
    - path_allele: Path object, the path to the fasta file.
    """
    # Read the content of the file.
    content = Path(path_allele).read_text()
    # Calculate the hash of the content.
    content_hash = hashlib.sha256(content.encode()).hexdigest()
    # Define the path to save the hash.
    path_cache_hash = Path(tempdir, "cache", "hash.txt")
    # Ensure the directory exists.
    path_cache_hash.parent.mkdir(parents=True, exist_ok=True)
    # Save the hash to the file.
    path_cache_hash.write_text(content_hash)



from __future__ import annotations

import random
import uuid
from collections import Counter, defaultdict
from itertools import groupby
from pathlib import Path
from typing import Generator

import cstag
import numpy as np
from rapidfuzz import process
from rapidfuzz.distance import DamerauLevenshtein
from sklearn.cluster import MeanShift

from DAJIN2.core.preprocess.mapping import to_sam
from DAJIN2.core.preprocess.sv_handler import add_unique_allele_keys, extract_unique_sv, save_cstag, save_fasta
from DAJIN2.utils import config, io
from DAJIN2.utils.cssplits_handler import convert_cssplit_to_dna, convert_cssplits_to_cstag

config.set_warnings_ignore()


###########################################################
# Cluster insertion alleles
###########################################################


def clustering_insertions(cssplits_insertion: list[str]) -> list[int]:
    seq_all = [convert_cssplit_to_dna(seq) for seq in cssplits_insertion]
    query = seq_all[0]
    _, distances, _ = zip(*process.extract_iter(query, seq_all, scorer=DamerauLevenshtein.normalized_distance))

    insertion_lengths = [[len(c) for c in cs.split(",")] for cs in cssplits_insertion]

    scores = [length + [distance] for length, distance in zip(insertion_lengths, distances)]

    return MeanShift(bin_seeding=True).fit_predict(np.array(scores)).tolist()


###########################################################
# Detect insertion sequences
###########################################################


def extract_all_insertions(midsv_sample: Generator, mutation_loci: list[set[str]]) -> dict[int, list[str]]:
    """To extract insertion sequences of **10 base pairs or more** at each index."""
    insertion_index_sequences_control = defaultdict(list)
    for m_sample in midsv_sample:
        cssplits = m_sample["CSSPLIT"].split(",")
        insertion_loci = (i for i, mut in enumerate(mutation_loci) if "+" in mut)
        for idx in insertion_loci:
            if cssplits[idx].startswith("+") and cssplits[idx].count("|") > 10:
                insertion_index_sequences_control[idx].append(cssplits[idx])

    return dict(insertion_index_sequences_control)


def extract_enriched_insertions(
    insertions_sample: dict, insertions_control: dict, coverage_sample: int
) -> dict[int, dict[str, int]]:
    enriched_insertions = {}
    threshold_sample = max(5, int(coverage_sample * 0.5 / 100))
    for i in insertions_sample:
        ins_sample: list[str] = insertions_sample[i]
        ins_control: list[str] = insertions_control.get(i, [])

        labels_all = clustering_insertions(ins_sample + ins_control)
        labels_sample = labels_all[: len(ins_sample)]
        labels_control = labels_all[len(ins_sample) : len(ins_sample) + len(ins_control)]

        labels_count_sample = dict(Counter(labels_sample))
        labels_count_control = dict(Counter(labels_control))

        to_delete = set()
        # To remove labels containing a high proportion of controls (5% or more, or 5 or more reads).
        threshold_control = max(5, int(len(labels_control) * 0.05))
        for label, count_control in labels_count_control.items():
            if count_control > threshold_control:
                to_delete.add(label)

        # To delete labels with sample coverage below 0.5% or fewer than 5 reads.
        for label, count_sample in labels_count_sample.items():
            if count_sample < threshold_sample:
                to_delete.add(label)

        for label in to_delete:
            if label in labels_count_sample:
                del labels_count_sample[label]

        if labels_count_sample == {}:
            continue

        # Count the remaining insertion sequences.
        enriched_insertions[i] = dict(
            Counter([ins for ins, label in zip(ins_sample, labels_sample) if label in labels_count_sample])
        )
    return enriched_insertions


def extract_insertions(
    path_sample: Path, path_control: Path, mutation_loci: list[set[str]]
) -> dict[int, dict[str, int]]:
    insertions_sample = extract_all_insertions(io.read_jsonl(path_sample), mutation_loci)
    insertions_control = extract_all_insertions(io.read_jsonl(path_control), mutation_loci)
    coverage_sample = io.count_newlines(path_sample)

    return extract_enriched_insertions(insertions_sample, insertions_control, coverage_sample)


###########################################################
# Merge similar insertion sequences
###########################################################


def group_index_by_consecutive_insertions(mutation_loci: list[set[str]]) -> list[tuple[int]]:
    """Groups indices of consecutive locations containing "+" mutations in the input list."""
    index = sorted(i for i, m in enumerate(mutation_loci) if "+" in m)
    index_grouped = []
    for _, group in groupby(enumerate(index), lambda i_x: i_x[0] - i_x[1]):
        items = [value for _, value in group]
        index_grouped.append(tuple(items))
    return index_grouped


def group_insertions(
    insertions: dict[int, dict[str, int]], index_grouped: list[tuple[int]]
) -> dict[tuple[int], dict[str, int]]:
    """
    Insertion bases in consecutive insertion base positions are grouped together in mutation_loci,
    as the insertion site may shift by a few bases.
    """
    insertions_grouped = defaultdict(dict)
    for idx in index_grouped:
        for i in idx:
            if i not in insertions:
                continue
            insertions_grouped[idx].update(insertions[i])
    return dict(insertions_grouped)


def get_merged_insertion(insertion: dict[str, int], labels: np.ndarray) -> dict[tuple[str], int]:
    insertion_label = [(label, {seq: count}) for label, (seq, count) in zip(labels, insertion.items())]
    insertion_label.sort(key=lambda x: x[0])

    insertion_merged = {}
    for _, group in groupby(insertion_label, key=lambda x: x[0]):
        group = [g[1] for g in group]
        sequences, counts = set(), 0
        for seq_count in group:
            seq, count = next(iter(seq_count.items()))
            sequences.add(seq)
            counts += count
        insertion_merged[tuple(sorted(sequences))] = counts

    return insertion_merged


def remove_minor_groups(
    insertions_merged: dict[tuple[int], dict[tuple[str], int]], coverage: int, percentage: float = 0.5
) -> dict[tuple[int], dict[tuple[str], int]]:
    """Remove minor groups with less than {percentage} % coverage or less than 5 reads."""
    threshold = max(5, int(coverage * percentage // 100))
    for _, ins in insertions_merged.items():
        # Create a list of elements to delete
        to_delete = []
        for seq, count in ins.items():
            if count < threshold:
                to_delete.append(seq)

        # Delete the collected elements
        for seq in to_delete:
            del ins[seq]

    return insertions_merged


def merge_similar_insertions(
    insertions: dict[int, dict[str, int]], mutation_loci: list[set[str]], coverage: int, percentage: float = 0.5
) -> dict[tuple[int], dict[tuple[str], int]]:
    index_grouped = group_index_by_consecutive_insertions(mutation_loci)
    insertions_grouped = group_insertions(insertions, index_grouped)
    insertions_merged = {}
    for idx, insertion in insertions_grouped.items():
        if len(insertion) == 1:
            seq, count = next(iter(insertion.items()))
            insertions_merged[idx] = {(seq,): count}
            continue

        labels = clustering_insertions(insertion)
        insertions_merged[idx] = get_merged_insertion(insertion, labels)

    return remove_minor_groups(insertions_merged, coverage, percentage)


###########################################################
# Pre- and post-process of clustering insertion alleles
###########################################################


def flatten_keys_to_set(dict_keys: list[tuple[int]]) -> set[int]:
    flattened_nums = set()
    for key_tuple in dict_keys:
        for num in key_tuple:
            flattened_nums.add(num)
    return flattened_nums


def subset_insertions(
    insertions_merged: dict[tuple[int], dict[tuple[str], int]], num_subset: int = 100
) -> dict[tuple[int], dict[tuple[str], int]]:
    """
    If the number of seqs exceeds `num_subset`, limit it to only the `num_subset` elements with fiexd random sampling.
    """
    random.seed(1)
    insertions_merged_subset = defaultdict(dict)
    for idx_grouped, insertions in insertions_merged.items():
        for cs_insertion, counts in insertions.items():
            if len(cs_insertion) > num_subset:
                sequences = random.sample(cs_insertion, num_subset)
            else:
                sequences = cs_insertion
            sequences = tuple(convert_cssplit_to_dna(seq) for seq in sequences)
            insertions_merged_subset[idx_grouped][sequences] = counts

    return dict(insertions_merged_subset)


def extract_score_and_sequence(
    path_sample, insertions_merged: dict[tuple[int], dict[tuple[str], int]]
) -> list[tuple[list[int], str]]:
    scores = []
    sequences = []

    insertions_merged_subset = subset_insertions(insertions_merged, num_subset=100)
    set_keys = flatten_keys_to_set(insertions_merged.keys())

    cache_score = defaultdict(int)
    for m in io.read_jsonl(path_sample):
        cssplits = m["CSSPLIT"].split(",")

        # Skip if there is no insertion
        if all(not cssplits[i].startswith("+") for i in set_keys):
            continue

        score = [0] * len(insertions_merged)
        sequence = ["N"] * len(insertions_merged)
        for i, (idx_grouped, insertions) in enumerate(insertions_merged.items()):
            for idx in idx_grouped:
                if not cssplits[idx].startswith("+"):
                    continue

                # Use the cache if there is already a precomputed value
                if (idx, cssplits[idx]) in cache_score:
                    score[i] = cache_score[(idx, cssplits[idx])]
                    sequence[i] = cssplits[idx]
                    continue

                # Record the insertion sequence and count if the sample and insertion sequences are the same
                for seqs, count in insertions.items():
                    if cssplits[idx] in seqs:
                        score[i] = count
                        sequence[i] = cssplits[idx]
                        cache_score[(idx, cssplits[idx])] = count
                        continue

                flag_break = False
                for seqs, count in insertions_merged_subset[idx_grouped].items():
                    for _, distance, _ in process.extract_iter(
                        convert_cssplit_to_dna(cssplits[idx]), seqs, scorer=DamerauLevenshtein.normalized_distance
                    ):
                        # Record the insertion sequence and count if the sample and insertion sequences are similar
                        if distance < 0.1:
                            score[i] = count
                            sequence[i] = cssplits[idx]
                            flag_break = True
                            break

                    if flag_break:
                        cache_score[(idx, cssplits[idx])] = count
                        break

        if any(score):
            scores.append(score)
            sequences.append(",".join(sequence))

    return list(zip(scores, sequences))


def filter_minor_label(
    labels: list[int],
    insertions_scores_sequences: list[tuple[list[int], str]],
    coverage: int,
    percentage: float = 0.5,
) -> tuple[list[int], list[str]]:
    threshold = max(5, int(coverage * percentage // 100))
    labels_, counts_ = np.unique(labels, return_counts=True)
    minor_labels = {label for label, count in zip(labels_, counts_) if count < threshold}
    index_minor_labels = {i for i, label in enumerate(labels) if label in minor_labels}
    labels_filtered = [label for i, label in enumerate(labels) if i not in index_minor_labels]
    score_seq_filterd = [
        score_seq for i, score_seq in enumerate(insertions_scores_sequences) if i not in index_minor_labels
    ]
    return labels_filtered, score_seq_filterd


###########################################################
# Call consensus
###########################################################


def subset_sequences(sequences, labels, num=1000) -> list[dict]:
    """
    Downsampling to the required number of sequences when there is a huge number of inserted sequences
    """
    sequences_subset = []
    tmp_sequences = []
    random.seed(1)
    for sequence, label in zip(sequences, labels):
        tmp_sequences.append({"CSSPLIT": sequence, "LABEL": label})
    tmp_sequences.sort(key=lambda x: x["LABEL"])
    for _, group in groupby(tmp_sequences, key=lambda x: x["LABEL"]):
        group = list(group)
        if len(group) > num:
            sequences_subset += random.sample(group, num)
        else:
            sequences_subset += group
    return sequences_subset


def get_cstag_position(sam_insertions: list[str]) -> tuple[list[str], list[int]]:
    cs_tags = []
    positions = []
    for sam in sam_insertions:
        if sam.startswith("@"):
            continue
        cs_tags.append(sam.split("\t")[-1])
        positions.append(int(sam.split("\t")[3]))
    return cs_tags, positions


def mapping_insertion(
    TEMPDIR: Path, SAMPLE_NAME: str, cs_transposed: list[str], consensus_length: int, uuid4: str
) -> Generator[str]:
    # Temporarily cache the reference sequence
    cs_insertion = next(cs for cs in cs_transposed if cs != "N" and len(cs) == consensus_length)
    ref_seq = cstag.to_sequence(convert_cssplits_to_cstag([cs_insertion]))

    path_reference = Path(TEMPDIR, SAMPLE_NAME, "fasta", f"reference-{uuid4}.fasta")
    Path(path_reference).write_text(f">reference_insertion\n{ref_seq}\n")

    # Temporarily cache the query sequences
    path_query = Path(TEMPDIR, SAMPLE_NAME, "fasta", f"query-{uuid4}.fasta")
    with Path(path_query).open("w") as file:
        for i, cs_insertion in enumerate(cs_transposed):
            que_seq = cstag.to_sequence(convert_cssplits_to_cstag([cs_insertion]))
            file.write(f">query_insertion_{i}\n{que_seq}\n")

    # Switch the mappy settings to either long-read or short-read, depending on the length of the reference sequence
    if len(ref_seq) > 500:
        sam_insertions = to_sam(path_reference, path_query)
    else:
        options = {"k": 3, "min_dp_score": 1, "min_chain_score": 1, "min_cnt": 1}
        sam_insertions = to_sam(path_reference, path_query, preset="sr", options=options)

    return sam_insertions


def generate_consensus_insertions(TEMPDIR: Path, SAMPLE_NAME: str, cssplits: list[list[str]]) -> str:
    consensus_insertion = []
    cssplits_transposed = (list(cs) for cs in zip(*cssplits))
    for cs_transposed in cssplits_transposed:
        if all(True if cs == "N" else False for cs in cs_transposed):
            consensus_insertion.append("N")
            continue

        count_N = sum(1 for cs in cs_transposed if cs == "N")

        count_cs = defaultdict(int)
        for cs in cs_transposed:
            if cs == "N":
                continue
            count_cs[len(cs)] += 1

        consensus_length = max(count_cs, key=count_cs.get)
        count_consensus = sum(1 for cs in cs_transposed if len(cs) == consensus_length)

        if count_N > count_consensus:
            consensus_insertion.append("N")
            continue

        sam_insertions = mapping_insertion(TEMPDIR, SAMPLE_NAME, cs_transposed, consensus_length, str(uuid.uuid4()))

        cs_tags, positions = get_cstag_position(sam_insertions)
        if cs_tags == []:  # When not mapped at all, append `N`.
            consensus_insertion.append("N")
        else:
            cs_insertion_consensus = cstag.to_sequence(cstag.consensus(cs_tags, positions))
            cs_last = Counter(cs.split("|")[-1] for cs in cs_transposed if len(cs) == consensus_length).most_common()[
                0
            ][0]
            cssplits_insertion_consensus = "|".join("+" + cs for cs in cs_insertion_consensus[:-1]) + "|" + cs_last
            consensus_insertion.append(cssplits_insertion_consensus)

    return ",".join(consensus_insertion)


def remove_all_n(cons_sequence: dict[int, str]) -> dict[int, str]:
    """Remove all `N` sequences."""
    cons_sequence_removed = {}
    for label, seq in cons_sequence.items():
        if all(True if s == "N" else False for s in seq.split(",")):
            continue
        cons_sequence_removed[label] = seq
    return cons_sequence_removed


def call_consensus_of_insertion(
    TEMPDIR: Path, SAMPLE_NAME: str, insertion_sequences_subset: list[dict]
) -> dict[int, str]:
    """Generate consensus cssplits."""
    consensus_insertion_cssplits = {}
    insertion_sequences_subset.sort(key=lambda x: x["LABEL"])
    for label, group in groupby(insertion_sequences_subset, key=lambda x: x["LABEL"]):
        cssplits = [cs["CSSPLIT"].split(",") for cs in group]
        consensus_insertion_cssplits[label] = generate_consensus_insertions(TEMPDIR, SAMPLE_NAME, cssplits)

    return remove_all_n(consensus_insertion_cssplits)


###########################################################
# Generate cstag and FASTA
###########################################################


def extract_index_of_insertions(
    insertions: dict[int, dict[str, int]], insertions_merged: dict[tuple[int], dict[tuple[str], int]]
) -> list[int]:
    """`insertions_merged` contains multiple surrounding indices for a single insertion allele. Among them, select the one index where the insertion allele is most frequent."""
    index_of_insertions = []
    max_idx = -1
    for idx_group in insertions_merged:
        max_val = -1
        for idx in idx_group:
            if idx not in insertions:
                continue
            if max_val < sum(insertions[idx].values()):
                max_val = sum(insertions[idx].values())
                max_idx = idx
        if max_idx != -1:
            index_of_insertions.append(max_idx)

    return index_of_insertions


def generate_cstag(
    consensus_of_insertions: dict[str, str], index_of_insertions: list[int], sequence: str
) -> dict[str, str]:
    cstag_insertions = {}
    for label, cons_seq in consensus_of_insertions.items():
        cons_seq = cons_seq.split(",")
        list_sequence = list(sequence)
        for idx, seq in zip(index_of_insertions, cons_seq):
            if seq == "N":
                continue
            list_sequence[idx] = convert_cssplits_to_cstag([seq]) + "="
        cstag_insertions[label] = "cs:Z:=" + "".join(list_sequence)

    return cstag_insertions


def generate_fasta(cstag_insertions: dict[str, str]) -> dict[str, str]:
    fasta_insertions = {}
    for label, cs_tag in cstag_insertions.items():
        fasta_insertions[label] = cstag.to_sequence(cs_tag)

    return fasta_insertions


###########################################################
# main
###########################################################


def remove_temporal_files(TEMPDIR: Path, SAMPLE_NAME: str) -> None:
    _ = [path.unlink() for path in Path(TEMPDIR, SAMPLE_NAME, "fasta").glob("reference-*.fasta")]
    _ = [path.unlink() for path in Path(TEMPDIR, SAMPLE_NAME, "fasta").glob("query-*.fasta")]


def detect_insertions(TEMPDIR, SAMPLE_NAME, CONTROL_NAME, FASTA_ALLELES) -> None:
    PATH_SAMPLE = Path(TEMPDIR, SAMPLE_NAME, "midsv", "control", f"{SAMPLE_NAME}.jsonl")
    PATH_CONTROL = Path(TEMPDIR, CONTROL_NAME, "midsv", "control", f"{CONTROL_NAME}.jsonl")
    MUTATION_LOCI = io.load_pickle(Path(TEMPDIR, SAMPLE_NAME, "mutation_loci", "control", "mutation_loci.pickle"))

    insertions = extract_insertions(PATH_SAMPLE, PATH_CONTROL, MUTATION_LOCI)
    if insertions == {}:
        """If there is no insertion, return None"""
        return None

    coverage: int = io.count_newlines(PATH_SAMPLE)

    insertions_merged = merge_similar_insertions(insertions, MUTATION_LOCI, coverage, percentage=0.5)
    if all(True if v == {} else False for v in insertions_merged.values()):
        """"If all the insertion alleles were minor, return None"""
        return None

    # Clustering similar insertion alleles
    insertions_scores_sequences = extract_score_and_sequence(PATH_SAMPLE, insertions_merged)
    cssplits_insertion = [cssplit for _, cssplit in insertions_scores_sequences]
    labels = clustering_insertions(cssplits_insertion)
    labels_filtered, insertion_scores_sequences_filtered = filter_minor_label(
        labels, insertions_scores_sequences, coverage, percentage=0.5
    )

    # Consensus calling
    insertion_sequences_subset = subset_sequences(
        [seq for _, seq in insertion_scores_sequences_filtered], labels_filtered, num=1000
    )
    consensus_of_insertions = call_consensus_of_insertion(TEMPDIR, SAMPLE_NAME, insertion_sequences_subset)
    if consensus_of_insertions == {}:
        """
        If there is no insertion sequence, return None
        It is possible when all insertion sequence annotated as `N` that is filtered out
        """
        remove_temporal_files(TEMPDIR, SAMPLE_NAME)
        return None

    # Format the insertion sequences as cstag and fasta
    index_of_insertions = extract_index_of_insertions(insertions, insertions_merged)
    cstag_insertions = generate_cstag(consensus_of_insertions, index_of_insertions, FASTA_ALLELES["control"])
    fasta_insertions = generate_fasta(cstag_insertions)
    fasta_insertions_unique = extract_unique_sv(fasta_insertions, FASTA_ALLELES)

    if fasta_insertions_unique == {}:
        remove_temporal_files(TEMPDIR, SAMPLE_NAME)
        return None

    # Update labels
    cstag_insertions_update = {key: cstag_insertions[key] for key in fasta_insertions_unique.keys()}
    cstag_insertions_update = add_unique_allele_keys(cstag_insertions_update, FASTA_ALLELES, "insertion")
    fasta_insertions_update = add_unique_allele_keys(fasta_insertions_unique, FASTA_ALLELES, "insertion")

    save_cstag(TEMPDIR, SAMPLE_NAME, cstag_insertions_update)
    save_fasta(TEMPDIR, SAMPLE_NAME, fasta_insertions_update)

    remove_temporal_files(TEMPDIR, SAMPLE_NAME)



from __future__ import annotations

import ssl
from urllib.request import urlopen


def fetch_html_without_verification(url: str) -> str:
    context = ssl._create_unverified_context()  # Create an SSL context that temporarily disables verification
    with urlopen(url, context=context, timeout=10) as response:
        return response.read().decode("utf-8").split("\n")


def fetch_seq_coordinates(genome: str, blat_url: str, seq: str) -> dict:
    url = f"{blat_url}?db={genome}&type=BLAT&userSeq={seq}"
    records = fetch_html_without_verification(url)
    matches = []
    for record in records:
        if "100.0%" not in record:
            continue
        record_trim = [r for r in record.split(" ") if r]
        if record_trim[-1] == str(len(seq)):
            matches = record_trim

    if not matches:
        raise ValueError(f"{seq[:60]}... is not found in {genome}")

    chrom, strand, start, end, _ = matches[-5:]
    return {"chrom": chrom, "strand": strand, "start": int(start), "end": int(end)}


def fetch_coordinates(genome_coordinates: dict, genome_urls: dict, seq: str) -> dict:
    genome = genome_coordinates["genome"]
    blat_url = genome_urls["blat"]

    seq_start, seq_end = seq[:1000], seq[-1000:]
    coordinate_start = fetch_seq_coordinates(genome, blat_url, seq_start)
    coordinate_end = fetch_seq_coordinates(genome, blat_url, seq_end)

    chromosome, strand = coordinate_start["chrom"], coordinate_start["strand"]
    if strand == "+":
        start, end = coordinate_start["start"], coordinate_end["end"]
    else:
        start, end = coordinate_end["start"], coordinate_start["end"]

    return {"genome": genome, "chrom": chromosome, "start": start, "end": end, "strand": strand}


def fetch_chromosome_size(genome_coordinates: dict, genome_urls: dict) -> int:
    chrom = genome_coordinates["chrom"]
    genome = genome_coordinates["genome"]
    url = f"{genome_urls['goldenpath']}/{genome}/bigZips/{genome}.chrom.sizes"

    records = fetch_html_without_verification(url)
    for record in records:
        chrom_name, size = record.split("\t")
        if chrom == chrom_name:
            return int(size)
    raise ValueError(f"Chromosome {chrom} size not found.")



"""Constrained k-means clustering
Bradley, Paul S., Kristin P. Bennett, and Ayhan Demiriz. "Constrained k-means clustering." Microsoft Research, Redmond 20.0 (2000): 0.

The implementation is modified from @kuga-qiita. (https://qiita.com/kuga-qiita/items/5588d5469f3268b7fd39#constrained-k-means-1)
"""

from __future__ import annotations

from functools import partial

import numpy as np
from scipy.spatial.distance import cdist


def random(
    X: np.ndarray, n_clusters: int, random_state: np.random.RandomState, **kwargs
) -> tuple[np.ndarray, np.ndarray]:
    n_samples, _ = X.shape
    indices = random_state.choice(n_samples, size=n_clusters, replace=True)
    centers = X[indices]
    return centers, indices


def kmeans_plusplus(X, n_clusters, random_state):
    n_samples, _ = X.shape
    centers = np.empty((n_clusters, X.shape[1]), dtype=X.dtype)
    indices = np.empty(n_clusters, dtype=int)

    center_id = random_state.choice(n_samples)
    centers[0] = X[center_id]
    indices[0] = center_id

    closest_dist_sq = np.full(n_samples, np.inf)
    for i in range(1, n_clusters):
        dist_sq = np.sum((X - centers[i - 1]) ** 2, axis=1)
        closest_dist_sq = np.minimum(closest_dist_sq, dist_sq)
        probs = closest_dist_sq / np.sum(closest_dist_sq)
        cumulative_probs = np.cumsum(probs)
        r = random_state.rand()
        center_id = np.searchsorted(cumulative_probs, r)
        centers[i] = X[center_id]
        indices[i] = center_id

    return centers, indices


class ConstrainedKMeans:
    def __init__(
        self,
        n_clusters: int,
        min_cluster_size: int,
        max_cluster_size: int | None = None,
        random_state: int = 0,
        init: str = "k-means++",
        n_init: int = 10,
        max_iter: int = 300,
        tol: float = 1e-4,
        **kwargs,
    ):
        self.valid(
            n_clusters,
            min_cluster_size,
            max_cluster_size,
            random_state,
            init,
            n_init,
            max_iter,
            tol,
        )
        self.n_clusters = n_clusters
        self.min_cluster_size = min_cluster_size
        self.max_cluster_size = max_cluster_size
        self.random_state = np.random.RandomState(random_state)

        if init == "random":
            self.init = partial(
                random,
                n_clusters=self.n_clusters,
                random_state=self.random_state,
                **kwargs,
            )
        elif init == "k-means++":
            self.init = partial(
                kmeans_plusplus,
                n_clusters=self.n_clusters,
                random_state=self.random_state,
                **kwargs,
            )

        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.kwargs = kwargs
        self.centers = None
        self.labels = None

    def get_lp_params(self, n_samples: int) -> tuple[np.ndarray, np.ndarray, np.ndarray, list[float]]:
        X_nodes = np.arange(n_samples)
        cluster_nodes = np.arange(n_samples, n_samples + self.n_clusters)
        artificial_demand_node = np.array([n_samples + self.n_clusters])
        start_nodes = np.concatenate([np.repeat(X_nodes, self.n_clusters), cluster_nodes])
        end_nodes = np.concatenate(
            [
                np.tile(cluster_nodes, n_samples),
                np.repeat(artificial_demand_node, self.n_clusters),
            ]
        )
        capacities = np.concatenate(
            [
                np.ones(self.n_clusters * n_samples),
                np.full(
                    self.n_clusters,
                    n_samples - self.n_clusters * self.min_cluster_size
                    if self.max_cluster_size is None
                    else self.max_cluster_size - self.min_cluster_size,
                ),
            ]
        )
        supplies = np.concatenate(
            [
                np.ones(n_samples),
                -1 * np.full(self.n_clusters, self.min_cluster_size),
                np.array([-n_samples + self.n_clusters * self.min_cluster_size]),
            ]
        ).tolist()
        return start_nodes, end_nodes, capacities, supplies

    def simulated_annealing(self, X: np.ndarray, centers: np.ndarray) -> np.ndarray:
        """
        Perform simulated annealing to optimize cluster assignments.

        This method uses simulated annealing to assign data points to clusters,
        allowing for probabilistic acceptance of worse assignments to escape local optima.

        Parameters:
        X (np.ndarray): The input data array of shape (n_samples, n_features).
        centers (np.ndarray): The initial cluster centers of shape (n_clusters, n_features).

        Returns:
        np.ndarray: A mask array of shape (n_samples, n_clusters) indicating cluster assignments.
        """
        n_samples, n_clusters = X.shape[0], self.n_clusters
        dist_sq = cdist(X, centers, "sqeuclidean")
        labels = np.argmin(dist_sq, axis=1)

        T = 1.0
        T_min = 0.0001
        alpha = 0.9

        while T > T_min:
            # Vectorize sample processing and compute in batches
            idx = np.random.randint(0, n_samples, size=1000)
            current_labels = labels[idx]
            new_labels = np.random.randint(0, n_clusters, size=1000)

            # Select only the samples where the new label differs from the current label
            mask = current_labels != new_labels
            idx = idx[mask]
            current_labels = current_labels[mask]
            new_labels = new_labels[mask]

            if len(idx) == 0:
                T *= alpha
                continue

            current_dists = dist_sq[idx, current_labels]
            new_dists = dist_sq[idx, new_labels]

            delta_E = new_dists - current_dists
            prob = np.exp(-delta_E / T)
            accept = (delta_E < 0) | (prob > np.random.rand(len(prob)))

            labels[idx[accept]] = new_labels[accept]

            T *= alpha

        mask = np.zeros((n_samples, n_clusters))
        mask[np.arange(n_samples), labels] = 1
        return mask

    def fit_once(self, X: np.ndarray, **kwargs) -> tuple[np.ndarray, float]:
        centers_, _ = self.init(X, **kwargs)
        dist_sq = cdist(X, centers_, "sqeuclidean")
        mask = self.simulated_annealing(X, centers_)
        inertia = np.sum(mask * dist_sq)
        return centers_, inertia

    def fit(self, X: np.ndarray) -> ConstrainedKMeans:
        self.valid_tr(X)
        if X.ndim == 1:
            X = X[:, np.newaxis]

        best_inertia, best_centers = None, None
        for _ in range(self.n_init):
            centers_, inertia = self.fit_once(X, **self.kwargs)
            if best_inertia is None or inertia < best_inertia:
                best_inertia = inertia
                best_centers = centers_
        self.centers = best_centers

        dist_sq = cdist(X, self.centers, "sqeuclidean")

        for i in range(self.max_iter):
            mask = self.simulated_annealing(X, self.centers)
            centers_ = np.dot(mask.T, X) / np.sum(mask, axis=0)[:, np.newaxis]
            dist_sq = cdist(X, centers_, "sqeuclidean")
            centers_squared_diff = np.sum((centers_ - self.centers) ** 2)
            self.centers = centers_
            iter_ = i
            if centers_squared_diff <= self.tol:
                break

        self.inertia = np.sum(mask * dist_sq)
        self.labels = np.argmax(mask, axis=-1)
        self.iter_ = iter_
        return self

    def fit_predict(self, X: np.ndarray) -> np.ndarray:
        return self.fit(X).labels

    def predict(self, X: np.ndarray) -> np.ndarray:
        self.valid_pr(X)
        if X.ndim == 1:
            X = X[:, np.newaxis]
        mask = self.simulated_annealing(X, self.centers)
        labels = np.argmax(mask, axis=1)
        return labels

    def valid(
        self,
        n_clusters: int,
        min_cluster_size: int,
        max_cluster_size: int | None,
        random_state: int,
        init: str,
        n_init: int,
        max_iter: int,
        tol: float,
    ) -> None:
        if not isinstance(n_clusters, int):
            raise TypeError("n_clusters must be an integer")
        if not isinstance(min_cluster_size, int):
            raise TypeError("min_cluster_size must be an integer")
        if not isinstance(max_cluster_size, int) and max_cluster_size is not None:
            raise TypeError("max_cluster_size must be an integer or None")
        if not isinstance(random_state, int) and random_state is not None:
            raise TypeError("random_state must be an integer")
        if init not in ["random", "k-means++"]:
            raise TypeError("init must be 'random' or 'k-means++'")
        if not isinstance(n_init, int):
            raise TypeError("n_init must be an integer")
        if not isinstance(max_iter, int):
            raise TypeError("max_iter must be an integer")
        if not isinstance(tol, (int, float)):
            raise TypeError("tol must be an integer or float")

        if n_clusters < 2:
            raise ValueError("n_clusters must be 2 or more")
        if min_cluster_size < 0:
            raise ValueError("min_cluster_size must be 0 or more")
        if isinstance(max_cluster_size, int):
            if max_cluster_size < min_cluster_size or max_cluster_size < 0:
                raise ValueError("max_cluster_size must be at least min_cluster_size")
        if isinstance(random_state, int):
            if random_state < 0:
                raise ValueError("random_state must be 0 or more")
        if n_init <= 0:
            raise ValueError("n_init must be 1 or more")
        if max_iter < 1:
            raise ValueError("max_iter must be 1 or more")
        if tol < 0:
            raise ValueError("tol must be 0 or more")

    def valid_tr(self, X: np.ndarray) -> None:
        if not isinstance(X, np.ndarray):
            raise TypeError("X must be numpy.ndarray")
        if X.ndim < 1 or X.ndim > 2:
            raise ValueError("X must be 1-D or 2-D")
        if X.shape[0] < self.n_clusters or X.shape[0] < self.n_clusters * self.min_cluster_size:
            raise ValueError("X must have at least n_clusters * min_cluster_size rows")
        if self.max_cluster_size is not None:
            if X.shape[0] > self.n_clusters * self.max_cluster_size:
                raise ValueError("X must have at most n_clusters * max_cluster_size rows")

    def valid_pr(self, X: np.ndarray) -> None:
        if self.centers is None or self.labels is None:
            raise RuntimeError("fit method hasn't been called")
        if not isinstance(X, np.ndarray):
            raise TypeError("X must be numpy.ndarray")
        if X.ndim < 1 or X.ndim > 2:
            raise ValueError("X must be 1-D or 2-D")
        if X.shape[0] < self.n_clusters or X.shape[0] < self.n_clusters * self.min_cluster_size:
            raise ValueError("X must have at least n_clusters * min_cluster_size rows")
        if self.max_cluster_size is not None:
            if X.shape[0] > self.n_clusters * self.max_cluster_size:
                raise ValueError("X must have at most n_clusters * max_cluster_size rows")
        if X.ndim == 2 and X.shape[-1] != self.centers.shape[-1]:
            raise ValueError("input shape does not match the centers shape")



from DAJIN2.core.preprocess.cache_checker import exists_cached_genome, exists_cached_hash
from DAJIN2.core.preprocess.directory_manager import create_report_directories, create_temporal_directories
from DAJIN2.core.preprocess.genome_fetcher import fetch_chromosome_size, fetch_coordinates
from DAJIN2.core.preprocess.input_formatter import format_inputs
from DAJIN2.core.preprocess.insertion_detector import detect_insertions
from DAJIN2.core.preprocess.inversion_detector import detect_inversions
from DAJIN2.core.preprocess.knockin_handler import extract_knockin_loci
from DAJIN2.core.preprocess.mapping import generate_sam
from DAJIN2.core.preprocess.midsv_caller import generate_midsv
from DAJIN2.core.preprocess.mutation_extractor import cache_mutation_loci



from __future__ import annotations

import csv
import os
import shutil
import socket
import webbrowser
from contextlib import closing, redirect_stderr
from pathlib import Path
from threading import Timer

from flask import Flask, render_template, request
from waitress import serve
from werkzeug.utils import secure_filename

from DAJIN2 import main
from DAJIN2.utils import config


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def upload_files(files):
    for file in files:
        path = Path(app.config["UPLOAD_FOLDER"], secure_filename(file.filename))
        file.save(path)


def get_path_of_uploaded_files(UPLOAD_FOLDER, files):
    paths = []
    for file in files:
        path = str(UPLOAD_FOLDER) + "/" + secure_filename(file.filename)
        paths.append(path)
    return paths


app = Flask(__name__)


@app.route("/")
def root_page():
    return render_template("index.html")


@app.route("/submit", methods=["POST"])
def submit():
    name = request.form.get("name")
    TEMPDIR = Path(config.TEMP_ROOT_DIR, name)
    if TEMPDIR.exists():
        shutil.rmtree(TEMPDIR)
    UPLOAD_FOLDER = Path(TEMPDIR, "upload")
    UPLOAD_FOLDER.mkdir(parents=True, exist_ok=True)
    app.config["UPLOAD_FOLDER"] = UPLOAD_FOLDER

    files = request.files.getlist("sample")
    upload_files(files)
    PATH_SAMPLE = get_path_of_uploaded_files(UPLOAD_FOLDER, files)

    files = request.files.getlist("control")
    upload_files(files)
    PATH_CONTROL = get_path_of_uploaded_files(UPLOAD_FOLDER, files)

    files = request.files.getlist("allele")
    upload_files(files)
    PATH_ALLELE = get_path_of_uploaded_files(UPLOAD_FOLDER, files)

    genome = request.form.get("genome")
    threads = request.form.get("threads")
    if threads is None:
        threads = 1

    data = []
    for sample in PATH_SAMPLE:
        row = {"sample": sample, "name": name, "control": PATH_CONTROL[0], "allele": PATH_ALLELE[0]}
        if genome:
            row["genome"] = genome
        data.append(row)

    output_path = Path("TEMPDIR", "upload", "batch.csv")
    output_path.parent.mkdir(parents=True, exist_ok=True)

    with open(output_path, mode="w", newline="") as file:
        writer = csv.DictWriter(file, fieldnames=data[0].keys())
        writer.writeheader()
        writer.writerows(data)

    arguments = {}
    arguments["file"] = str(Path(TEMPDIR, "upload", "batch.csv"))
    arguments["threads"] = threads
    arguments["debug"] = False

    main.execute_batch_mode(arguments)

    return f"""
    name={name}
    sample={PATH_SAMPLE}
    control={PATH_CONTROL}
    allele={PATH_ALLELE}
    genome={genome}
    threads={threads}
    arguments={arguments["file"]}
    """


def open_browser(PORT):
    webbrowser.open_new(f"http://localhost:{PORT}/")


def execute():
    PORT = find_free_port()
    print(f"Assess 'http://localhost:{PORT}/' if a browser does not automatically open.")
    Timer(1, open_browser, [PORT]).start()
    with redirect_stderr(open(os.devnull, "w")):
        serve(app, host="0.0.0.0", port=PORT)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from itertools import groupby
from pathlib import Path

from DAJIN2.utils import io
from DAJIN2.utils.cssplits_handler import call_sequence

###########################################################
# call position weight matrix (cons_pergentage)
###########################################################


def convert_to_percentage(
    cssplits: list[list[str]], mutation_loci: list[set[str]], sequence: str
) -> list[dict[str, float]]:
    """
    Convert sequences and mutations into percentages, annotating sequence errors as "SEQERROR".
    """
    # Transpose the cssplits to facilitate per-position processing
    cssplits_transposed = [list(cs) for cs in zip(*cssplits)]
    coverage = len(cssplits_transposed[0])

    cons_percentage = []
    for cs_transposed, mut_loci, nucleotide in zip(cssplits_transposed, mutation_loci, sequence):
        count_cs = defaultdict(float)
        for cs in cs_transposed:
            operator = cs[0]
            if operator in {"+", "-", "*"}:
                if operator not in mut_loci:
                    cs = f"={nucleotide.upper()}"
            count_cs[cs] += 1 / coverage * 100
        cons_percentage.append(dict(count_cs))

    return cons_percentage


def remove_all_n(cons_percentage: list[dict[str, float]]) -> list[dict[str, float]]:
    for c in cons_percentage:
        if len(c) == 1 and "N" in c:
            continue
        c.pop("N", None)

    return cons_percentage


def adjust_to_100_percent(cons_percentage: list[dict[str, float]]) -> list[dict[str, float]]:
    adjusted_percentages = []

    for percentage_dict in cons_percentage:
        total = sum(percentage_dict.values())
        scaling_factor = 100 / total

        adjusted_dict = {key: value * scaling_factor for key, value in percentage_dict.items()}
        adjusted_percentages.append(adjusted_dict)

    return adjusted_percentages


def call_percentage(cssplits: list[list[str]], mutation_loci: list[set[str]], sequence: str) -> list[dict[str, float]]:
    """Call position weight matrix in different loci. Non-different loci are annotated as "Match"."""
    cons_percentage = convert_to_percentage(cssplits, mutation_loci, sequence)
    cons_percentage = remove_all_n(cons_percentage)
    return adjust_to_100_percent(cons_percentage)


###########################################################
# main
###########################################################


@dataclass(frozen=True)
class ConsensusKey:
    allele: str
    label: int
    percent: float


def call_consensus(
    tempdir: Path, sample_name: str, fasta_alleles: dict[str, str], clust_sample: list[dict]
) -> tuple[dict[list], dict[str]]:
    clust_sample.sort(key=lambda x: [x["ALLELE"], x["LABEL"]])

    cons_percentages = {}
    cons_sequences = {}

    for (allele, label), group in groupby(clust_sample, key=lambda x: [x["ALLELE"], x["LABEL"]]):
        clust = list(group)
        sequence = fasta_alleles[allele]

        path_consensus = Path(tempdir, sample_name, "consensus", allele, str(label))
        cons_mutation_loci = io.load_pickle(Path(path_consensus, "mutation_loci.pickle"))

        cssplits = [cs["CSSPLIT"].split(",") for cs in clust]
        cons_percentage = call_percentage(cssplits, cons_mutation_loci, sequence)

        key = ConsensusKey(allele, label, clust[0]["PERCENT"])
        cons_percentages[key] = cons_percentage
        cons_sequences[key] = call_sequence(cons_percentage)
    return cons_percentages, cons_sequences



from __future__ import annotations

from itertools import groupby
from pathlib import Path

from DAJIN2.core.classification.allele_merger import merge_minor_alleles
from DAJIN2.utils import io


def calc_match(cssplit: str) -> int:
    match_score = cssplit.count("=")
    match_score -= cssplit.count("+")  # insertion
    match_score -= cssplit.count("-")  # deletion
    match_score -= sum(cs.islower() for cs in cssplit)  # inversion

    return match_score


def score_allele(path_midsv: Path, allele: str) -> list[dict]:
    midsv_sample = io.read_jsonl(path_midsv)
    scored_alleles = []
    for dict_midsv in midsv_sample:
        score = calc_match(dict_midsv["CSSPLIT"])
        dict_midsv.update({"SCORE": score, "ALLELE": allele})
        scored_alleles.append(dict_midsv)

    return scored_alleles


def extract_alleles_with_max_score(score_of_each_alleles: list[dict]) -> list[dict]:
    alleles_with_max_score = []
    score_of_each_alleles.sort(key=lambda x: x["QNAME"])
    for _, group in groupby(score_of_each_alleles, key=lambda x: x["QNAME"]):
        max_read = max(group, key=lambda x: x["SCORE"])
        del max_read["SCORE"]
        alleles_with_max_score.append(max_read)

    return alleles_with_max_score


##########################################################
# main
##########################################################


def classify_alleles(TEMPDIR: Path, FASTA_ALLELES: dict, SAMPLE_NAME: str) -> list[dict]:
    score_of_each_alleles = []
    for allele in FASTA_ALLELES:
        path_midsv = Path(TEMPDIR, SAMPLE_NAME, "midsv", allele, f"{SAMPLE_NAME}.jsonl")
        score_of_each_alleles.extend(score_allele(path_midsv, allele))

    score_of_each_alleles_merged = merge_minor_alleles(score_of_each_alleles, threshold=5)

    return extract_alleles_with_max_score(score_of_each_alleles_merged)



from __future__ import annotations

import gzip
import random
import re
import uuid
from pathlib import Path

import mappy
import pysam
from DAJIN2.utils.io import sanitize_name

#################################################
# Extract filename
#################################################


def extract_filename(path_fasta: Path | str) -> str:
    filename = Path(path_fasta).name
    filename = re.sub(r"\..*$", "", filename)  # Remove file extension
    return sanitize_name(filename)


#################################################
# Convert allele file to dictionary type fasta format
#################################################


def dictionize_allele(path_fasta: str | Path) -> dict[str, str]:
    return {sanitize_name(name): seq.upper() for name, seq, _ in mappy.fastx_read(str(path_fasta))}


#################################################
# Export fasta files as single-FASTA format
#################################################


def export_fasta_files(TEMPDIR: Path, FASTA_ALLELES: dict, NAME: str) -> None:
    """Save multiple FASTAs in separate single-FASTA format files."""
    for identifier, sequence in FASTA_ALLELES.items():
        identifier = sanitize_name(identifier)
        contents = "\n".join([">" + identifier, sequence]) + "\n"
        path_output_fasta = Path(TEMPDIR, NAME, "fasta", f"{identifier}.fasta")
        path_output_fasta.write_text(contents)


#################################################
# save_concatenated_fastx
#################################################


def extract_extention(path_file: str | Path) -> str:
    suffixes = Path(path_file).suffixes
    return "".join(suffixes)


def is_gzip_file(path_file: str | Path) -> bool:
    """Check if a file is a GZip compressed file."""
    try:
        with Path(path_file).open("rb") as f:
            return f.read(2) == b"\x1f\x8b"
    except OSError:
        return False


def save_bam_files_to_single_fastq(TEMPDIR: Path, path_bam_files: list[Path], sample_name: str) -> None:
    path_bam_files = [str(path) for path in path_bam_files]
    path_concatenated_bam = Path(TEMPDIR, sample_name, "fastq", f"tmp_{sample_name}_{str(uuid.uuid4())}.bam")
    pysam.merge("-f", "-o", str(path_concatenated_bam), *path_bam_files)

    path_concatenated_fastq = Path(TEMPDIR, sample_name, "fastq", f"{sample_name}.fastq.gz")
    _ = pysam.fastq("-0", str(path_concatenated_fastq), str(path_concatenated_bam))

    path_concatenated_bam.unlink()


def save_fastx_files_to_single_fastq(TEMPDIR: Path, path_input_files: list[Path], sample_name: str) -> None:
    """Merge gzip and non-gzip files into a single gzip file."""
    with gzip.open(Path(TEMPDIR, sample_name, "fastq", f"{sample_name}.fastq.gz"), "wb") as merged_file:
        for path_file in path_input_files:
            if is_gzip_file(path_file):
                with gzip.open(path_file, "rb") as f:
                    merged_file.write(f.read())
            else:
                with open(path_file) as f:
                    merged_file.write(f.read().encode())


def save_inputs_as_single_fastx(TEMPDIR: Path, path_directory: Path, sample_name: str) -> None:
    file_suffix = {".fa", ".fq", ".fasta", ".fastq", ".fa.gz", ".fq.gz", ".fasta.gz", ".fastq.gz", ".bam"}
    path_input_files = [path for path in path_directory.iterdir() if extract_extention(path) in file_suffix]
    if all(extract_extention(path) == ".bam" for path in path_directory.iterdir()):
        save_bam_files_to_single_fastq(TEMPDIR, path_input_files, sample_name)
    else:
        save_fastx_files_to_single_fastq(TEMPDIR, path_input_files, sample_name)


#################################################
# save_subsetted_fastx
#################################################


def read_lines(path_file) -> list[str]:
    if is_gzip_file(path_file):
        with gzip.open(path_file, "rt") as f:
            return [line.strip() for line in f]
    else:
        with open(path_file) as f:
            return [line.strip() for line in f]


def parse_fastq(fastq) -> list[dict]:
    fastq_parsed = []
    iterator = iter(fastq)
    try:
        while True:
            header = next(iterator).strip()
            sequence = next(iterator).strip()
            annotate = next(iterator).strip()
            quality = next(iterator).strip()
            fastq_parsed.append({"header": header, "sequence": sequence, "annotate": annotate, "quality": quality})
    except StopIteration:
        pass
    return fastq_parsed


def save_subsetted_fastx(path_fastq: str | Path, num_reads: int = 10_000) -> None:
    """If the number of control reads is too high, it unnecessarily slows down the computation speed. Therefore, we perform random sampling to reduce the number of reads to below 10,000."""

    fastq: list[str] = read_lines(path_fastq)
    if len(fastq) // 4 <= 10_000:
        return None

    fastq: list[dict] = parse_fastq(fastq)

    random.seed(1)
    fastq_subset = random.sample(fastq, num_reads)
    with gzip.open(path_fastq, "wt") as f:
        for read in fastq_subset:
            f.write(f"{read['header']}\n{read['sequence']}\n{read['annotate']}\n{read['quality']}\n")



from __future__ import annotations

import re

import numpy as np


def get_repeat_regions(sequence: str, loci: set[int]) -> list[tuple[int, int]]:
    """
    Find homopolymers in the sequence but discard them that
    are adjacent to candidate mutation loci because they are
    likely to be covered by the real mutations
    """
    pattern = r"A{4,}|C{4,}|G{4,}|T{4,}|N{4,}"
    repeat_regions = []
    for start, end in (match.span() for match in re.finditer(pattern, sequence)):
        if not (start - 1 in loci and end + 1 in loci):
            repeat_regions.append((start, end))
    return repeat_regions


def cosine_similarity(X, Y) -> float:
    # Add 1e-6 to avoid division by zero when calculating cosine similarity
    X += 1e-6
    Y += 1e-6
    return float(np.dot(X, Y) / (np.linalg.norm(X) * np.linalg.norm(Y)))


###########################################################
# main
###########################################################


def extract_sequence_errors_in_homopolymer_loci(
    sequence: str,
    indels_normalized_sample: dict[str, np.array],
    indels_normalized_control: dict[str, np.array],
    anomal_loci: dict[set],
) -> dict[str, set[int]]:
    sequence_errors_in_homopolymer = {}
    for mut in ["+", "-", "*"]:
        repeat_regions = get_repeat_regions(sequence, anomal_loci[mut])
        if len(repeat_regions) == 0:
            sequence_errors_in_homopolymer[mut] = set()
            continue
        sequence_errors = set()
        for start, end in repeat_regions:
            x = np.array(indels_normalized_sample[mut][start:end])
            y = np.array(indels_normalized_control[mut][start:end])

            # Scaling data to [0, 1] for cosine similarity

            # Check if the range of x is zero
            if x.max() - x.min() == 0:
                x_scaled = np.zeros_like(x)
            else:
                x_scaled = (x - x.min()) / (x.max() - x.min())

            # Check if the range of y is zero
            if y.max() - y.min() == 0:
                y_scaled = np.zeros_like(y)
            else:
                y_scaled = (y - y.min()) / (y.max() - y.min())

            if cosine_similarity(x_scaled, y_scaled) > 0.95:
                sequence_errors.update(range(start, end + 1))

        sequence_errors_in_homopolymer[mut] = sequence_errors

    return sequence_errors_in_homopolymer



from __future__ import annotations

import random
from collections import defaultdict
from itertools import groupby


def remove_minor_alleles(clust_sample: list[dict]) -> list[dict]:
    """Remove minor alleles from the clustering sample."""
    min_sample_size = max(5, int(len(clust_sample) * 0.5 // 100))

    counts = defaultdict(int)
    for c in clust_sample:
        counts[c["LABEL"]] += 1

    minor_labels = {label for label, count in counts.items() if count < min_sample_size}

    return [c for c in clust_sample if c["LABEL"] not in minor_labels]


def downsample_by_label(clust_sample: list[dict], num: int = 1000) -> list[dict]:
    random.seed(1)

    clust_downsampled = []
    clust_sample.sort(key=lambda x: x["LABEL"])
    for _, group in groupby(clust_sample, key=lambda x: x["LABEL"]):
        samples = list(group)
        num_samples = min(len(samples), num)
        clust_downsampled.extend(random.sample(samples, num_samples))

    return clust_downsampled



from __future__ import annotations

import re

import cstag
import numpy as np
import ruptures as rpt


def find_n_boundaries(cssplits: list[str]) -> tuple[int, int]:
    """Find the boundaries of contiguous Ns which aren't at the ends."""

    # Find the left boundary
    left_idx_n = 0
    for char in cssplits:
        if char != "N":
            break
        left_idx_n += 1

    # Find the right boundary
    right_idx_n = len(cssplits) - 1
    for char in reversed(cssplits):
        if char != "N":
            break
        right_idx_n -= 1

    return left_idx_n - 1, right_idx_n + 1

# ToDo: 点変異が扱えていない。convert_cssplits_to_cstagをinversion検出用に改変すれば、必要なくなる。
def convert_cssplit_to_dna(cssplit: str) -> str:
    """Convert a cssplit to a plain DNA sequence."""
    return "".join([char for char in cssplit if char.isalpha()])


###########################################################
# reverse complement to cssplits
###########################################################


def _reverse_cssplits(cssplits: list[str]) -> list[str]:
    for i, cs in enumerate(cssplits):
        if cs.startswith("+"):
            cssplits[i] = "+" + "|".join(cs.split("|")[::-1])
    return cssplits[::-1]


def _realign_insertion(cssplits: list[str]) -> list[str]:
    for i, cs in enumerate(cssplits):
        if not cs.startswith("+"):
            continue
        if re.search(rf"[{cs[1]}]", "[ACGTacgt]"):
            continue
        if i + 1 == len(cssplits):
            continue
        cs_current = cs.split("|")
        cssplits[i] = cs_current[0].replace("+", "")
        cssplits[i + 1] = "|".join(c[0] + c[-1] for c in cs_current[1:]) + "|" + cssplits[i + 1]
    return cssplits


def _complement_cssplit(cssplits: list[str]) -> list[str]:
    comp = {"A": "T", "C": "G", "G": "C", "T": "A", "N": "N", "a": "t", "c": "g", "g": "c", "t": "a", "n": "n"}
    for i, cs in enumerate(cssplits):
        op = cs[0]
        if op == "*":
            cssplits[i] = op + comp[cs[1]] + comp[cs[2]]
        elif op == "+":
            cssplits[i] = "|".join(c[0] + comp[c[-1]] for c in cs.split("|"))
        else:  # Match or Deletion or N
            cssplits[i] = op + comp[cs[-1]]
        cssplits[i] = cssplits[i].replace("NN", "N")
        cssplits[i] = cssplits[i].replace("nn", "n")
    return cssplits


def revcomp_cssplits(cssplits: list[str]) -> list[str]:
    cssplits_reversed = _reverse_cssplits(cssplits)
    cssplits_realigned = _realign_insertion(cssplits_reversed)
    cssplits_revcomped = _complement_cssplit(cssplits_realigned)
    return cssplits_revcomped


###########################################################
# convert cssplits to cstag
###########################################################


def _add_match_operator_to_n(cssplits: list[str]) -> list[str]:
    """Add "=" (match operator) to the sequences with "N"."""
    cssplits_add_match_op = []
    for cs in cssplits:
        if cs.startswith("N") or cs.startswith("n"):
            cssplits_add_match_op.append("=" + cs)
        elif cs.startswith("+") and (cs[-1] == "N" or cs[-1] == "n"):
            cs_ins = cs.split("|")
            cs_last = "=" + cs_ins[-1]
            cs = "|".join(cs_ins[:-1]) + "|" + cs_last
            cssplits_add_match_op.append(cs)
        else:
            cssplits_add_match_op.append(cs)
    return cssplits_add_match_op


def _split_cssplits_by_delimiter(cssplits: list[str]) -> list[str]:
    cssplits_break = []
    for cs in cssplits:
        if cs.startswith("+"):
            cssplits_break.extend(cs.split("|"))
        else:
            cssplits_break.append(cs)
    return cssplits_break


def _combine_cssplits_by_prefix(cssplits: list[str]) -> list[str]:
    if not cssplits:
        return []

    combined_cssplits = []
    prev_prefix: str = cssplits[0][0]
    current_cs: list[str] = [cssplits[0][1:]]

    for item in cssplits[1:]:
        current_prefix, cs = item[0], item[1:]
        if prev_prefix == current_prefix:
            if current_prefix == "*":
                current_cs.append(item)
            else:
                current_cs.append(cs)
        else:
            combined_cssplits.append(prev_prefix + "".join(current_cs))
            prev_prefix = current_prefix
            current_cs = [cs]

    combined_cssplits.append(prev_prefix + "".join(current_cs))
    return combined_cssplits


def _standardize_case(cssplits: list[str]) -> list[str]:
    """Standardize the case of characters based on mutation types."""
    transformations = {"*": str.lower, "-": str.lower, "+": str.lower, "~": str.lower, "=": str.upper}

    transformed = []
    for cs in cssplits:
        prefix = cs[0]
        trans_func = transformations.get(prefix)
        transformed.append(prefix + trans_func(cs[1:]))

    return transformed


def convert_cssplits_to_cstag(cssplits: list[str]) -> str:
    cssplits_matched_op = _add_match_operator_to_n(cssplits)
    cssplits_splitted = _split_cssplits_by_delimiter(cssplits_matched_op)
    cssplits_combined = _combine_cssplits_by_prefix(cssplits_splitted)
    return "".join(_standardize_case(cssplits_combined))


def call_sequence(cons_percentage: list[dict[str, float]]) -> str:
    """convert position weight matrix (cons_pergentage) to sequence"""

    consensus_sequence = []
    for cons_per in cons_percentage:
        cssplits = max(cons_per, key=cons_per.get)
        cs_tag = convert_cssplits_to_cstag([cssplits])
        seq = cstag.to_sequence(cs_tag)
        consensus_sequence.append(seq)
    return "".join(consensus_sequence)


###########################################################
# reallocate_insertion_within_deletion
###########################################################


def _extract_candidate_index_of_large_deletions(
    cssplits: list[str], bin_size: int = 500, percentage: int = 50
) -> list[dict[str, int]]:
    deletion_counts = []
    for start in range(0, len(cssplits), bin_size // 10):
        end = start + bin_size
        counts_deletion = "".join(cssplits[start:end]).count("-")
        deletion_counts.append((start, counts_deletion))

    range_of_large_deletions = []
    on_the_peak = False
    start = -1
    end = -1
    for i, count in deletion_counts:
        del_percentage = count / bin_size * 100
        if del_percentage > percentage and on_the_peak is False:
            start = i
            on_the_peak = True
        if del_percentage <= percentage and on_the_peak is True:
            end = i
            on_the_peak = False
        if start != -1 and end != -1:
            range_of_large_deletions.append({"start": start, "end": end})
            start = -1
            end = -1

    return range_of_large_deletions


def _extract_break_points_of_large_deletions(
    cssplits: list[str], range_of_large_deletions: list[dict[str, int]], bin_size: int = 500
):
    cssplits_match_bool = np.array([1 if cs.startswith("=") else 0 for cs in cssplits], dtype=bool)

    break_points = []
    for range_of_deletion in range_of_large_deletions:
        break_point = {}
        for key, index in range_of_deletion.items():
            window_start = index
            window_end = index + bin_size

            # Find breakpoints(bps). First, predict approximate breakpoints using the signal of the window.
            signal = cssplits_match_bool[window_start:window_end]
            bps = rpt.Window(model="ar").fit(signal).predict(n_bkps=10)[:-1]

            # Aling breakpoints to the original index
            if bps:
                approximate_bp: int = bps[0] if key == "start" else bps[-1]
            else:
                approximate_bp: int = 0 if key == "start" else bin_size

            # Find the exact breakpoint because the approximate breakpoint may not be accurate
            signal_subset = signal[max(0, approximate_bp - 50) : min(approximate_bp + 50, len(signal))]
            bp = -1
            for i, sig in enumerate(signal_subset):
                if key == "start" and not sig:
                    bp = i + max(0, approximate_bp - 50) + window_start
                    break
                if key == "end" and sig:
                    bp = i + max(0, approximate_bp - 50) + window_start - 1
                    break
            if bp == -1:
                bp = approximate_bp

            # Store the breakpoint
            break_point[key] = bp

        if break_point["start"] == break_point["end"]:
            continue

        break_points.append(break_point)

    return break_points


def _convert_break_points_to_index(break_points: list[dict[str, int]]) -> list[int]:
    index_of_large_deletions = []
    for break_point in break_points:
        start = break_point["start"]
        end = break_point["end"]
        index_of_large_deletions += list(range(start, end + 1))

    return index_of_large_deletions


def _find_matched_indexes(cssplits: list[str], index_of_large_deletions: list[int]) -> list[int]:
    matched_index = []
    count_matches = 0
    start_match = -1

    index_of_large_deletions.sort()
    for i in index_of_large_deletions:
        if cssplits[i].startswith("="):
            if start_match == -1:
                start_match = i
            count_matches += 1
        else:
            if count_matches >= 10:
                matched_index += list(range(start_match, i))
            count_matches = 0
            start_match = -1

    return matched_index


def _remove_matched_indexes(index_of_large_deletions: list[int], matched_index: list[int]) -> set[int]:
    return set(index_of_large_deletions) - set(matched_index)


def _get_index_of_large_deletions(cssplits: list[str], bin_size: int = 500, percentage: int = 50) -> set[int]:
    range_of_large_deletions = _extract_candidate_index_of_large_deletions(cssplits, bin_size, percentage)
    break_points = _extract_break_points_of_large_deletions(cssplits, range_of_large_deletions, bin_size)

    index_of_large_deletions = _convert_break_points_to_index(break_points)
    matched_index = _find_matched_indexes(cssplits, index_of_large_deletions)
    return _remove_matched_indexes(index_of_large_deletions, matched_index)


def _adjust_cs_insertion(cs: str) -> str:
    if cs.startswith("-"):
        return None
    elif cs.startswith("+"):
        ins = "|".join(cs.split("|")[:-1])
        end_cs = cs.split("|")[-1][-1]
        cs_insertion = f"{ins}|+{end_cs}|"
    else:
        cs_insertion = f"+{cs[-1]}|"
    return cs_insertion


def reallocate_insertion_within_deletion(cssplits: list[str], bin_size: int = 500, percentage: int = 50) -> list[str]:
    """
    Since the mapping in minimap2 is local alignment, insertion bases within large deletions may be partially mapped to the reference genome and not detected as insertion bases. Therefore, update cssplits to detect insertions within large deletions as insertions.
    """
    cssplits_updated = cssplits.copy()

    index_of_large_deletions: set[int] = _get_index_of_large_deletions(
        cssplits, bin_size=bin_size, percentage=percentage
    )

    insertion_within_deletion = []
    for i, cs in enumerate(cssplits):
        if i in index_of_large_deletions:
            if cs.startswith("-"):
                continue
            if cs.startswith("N") or cs.startswith("n"):
                cs_new = cs
            elif cs.startswith("*"):
                cs_new = f"-{cs[1]}"
            elif cs.startswith("+") and cs.split("|")[-1].startswith("*"):
                cs_new = f'-{cs.split("|")[-1][1]}'
            else:
                cs_new = f"-{cs[-1]}"
            cssplits_updated[i] = cs_new

            insertion_within_deletion.append(_adjust_cs_insertion(cs))
        else:
            if not insertion_within_deletion:
                continue
            cssplits_updated[i] = "".join(insertion_within_deletion) + cs
            insertion_within_deletion = []

    return cssplits_updated



from __future__ import annotations

import textwrap
from pathlib import Path

import cstag

from DAJIN2.core.report.insertion_reflector import reflect_ref_insertion_to_query
from DAJIN2.utils.cssplits_handler import convert_cssplits_to_cstag, reallocate_insertion_within_deletion


def convert_to_fasta(header: str, sequence: str) -> str:
    header = ">" + header
    sequence_wrapped = textwrap.wrap(sequence, 80)
    fasta = "\n".join([header, *sequence_wrapped]) + "\n"

    return fasta


def convert_to_html(TEMPDIR: Path, SAMPLE_NAME: str, header: str, cons_per: list[dict]) -> str:
    cons_cssplits = [max(cons, key=cons.get) for cons in cons_per]
    cons_cssplits_reallocated = reallocate_insertion_within_deletion(cons_cssplits, bin_size=500, percentage=50)
    cons_cstag = convert_cssplits_to_cstag(cons_cssplits_reallocated)

    allele = header.split("_")[1]
    if Path(TEMPDIR, SAMPLE_NAME, "cstag", f"{allele}.txt").exists():
        ref_cstag = Path(TEMPDIR, SAMPLE_NAME, "cstag", f"{allele}.txt").read_text()
        cons_cstag = reflect_ref_insertion_to_query(ref_cstag, cons_cstag)

    return cstag.to_html(cons_cstag, f"{SAMPLE_NAME} {header.replace('_', ' ')}")


##################################################
# Export files
##################################################


def export_to_fasta(TEMPDIR: Path, SAMPLE_NAME: str, cons_sequence: dict) -> None:
    for header, sequence in cons_sequence.items():
        path_output = Path(TEMPDIR, "report", "FASTA", SAMPLE_NAME, f"{SAMPLE_NAME}_{header}.fasta")
        path_output.write_text(convert_to_fasta(f"{SAMPLE_NAME}_{header}", sequence))


def parse_fasta(file_path: Path) -> tuple[str, str]:
    """Parses a FASTA file and returns the header and concatenated sequence."""
    with open(file_path) as f:
        lines = f.readlines()

    header = lines[0].strip().lstrip(">")
    sequence = "".join(line.strip() for line in lines[1:])

    return header, sequence


def export_reference_to_fasta(TEMPDIR: Path, SAMPLE_NAME: str) -> None:
    for fasta in Path(TEMPDIR, SAMPLE_NAME, "fasta").glob("*.fasta"):
        header, sequence = parse_fasta(fasta)
        path_output = Path(TEMPDIR, "report", "FASTA", SAMPLE_NAME, f"{header}.fasta")
        path_output.write_text(convert_to_fasta(f"{SAMPLE_NAME}_{header}", sequence))


def export_to_html(TEMPDIR: Path, SAMPLE_NAME: str, cons_percentage: dict[list]) -> None:
    for header, cons_per in cons_percentage.items():
        path_output = Path(TEMPDIR, "report", "HTML", SAMPLE_NAME, f"{SAMPLE_NAME}_{header}.html")
        path_output.write_text(convert_to_html(TEMPDIR, SAMPLE_NAME, header, cons_per))


# def to_vcf(TEMPDIR: Path, SAMPLE_NAME: str, GENOME_COODINATES: dict[str, str], cons_percentage: dict[list]) -> str:
#     pass
#     for header, cons_per in cons_percentage.items():
#         path_output = Path(TEMPDIR, "report", "HTML", SAMPLE_NAME, f"{SAMPLE_NAME}_{header}.vcf")
#         path_output.write_text(_to_html(TEMPDIR, SAMPLE_NAME, header, cons_per))



from __future__ import annotations

import io
import logging
import sys
import traceback
from collections.abc import Callable
from itertools import islice
from multiprocessing import Process, Queue
from typing import Generator


def get_error_message_prefix(arg: dict) -> str:
    """Generate the error message prefix based on the argument's sample value."""
    return f"An unexpected error occurred at {arg['sample']}"


def generate_chunks(iterable, chunk_size: int = 1) -> Generator[tuple]:
    """Yield successive n-sized chunks from iterable."""
    iterator = iter(iterable)
    while True:
        chunk = tuple(islice(iterator, chunk_size))
        if not chunk:
            return
        yield chunk


def suppress_stderr(function: Callable, *args, **kwargs) -> None:
    """
    Suppress stderr and execute the function.
    If an error occurs, revert the stderr and raise the exception.
    """
    original_stderr = sys.stderr
    sys.stderr = io.StringIO()

    try:
        function(*args, **kwargs)
    finally:
        sys.stderr = original_stderr


def handle_exception_and_enqueue(queue: Queue[str], arg: dict) -> None:
    """
    Handle the exception by placing an error message in the queue.
    """
    error_message = f"{get_error_message_prefix(arg)}\n{traceback.format_exc()}"
    queue.put(error_message)


def target(function: Callable, arg: dict, queue: Queue[str]) -> None:
    """Run a function with a single argument and handle any exceptions."""
    try:
        suppress_stderr(function, arg)
    except Exception:
        handle_exception_and_enqueue(queue, arg)
        raise


def run(function: Callable, arguments: list[dict], num_workers: int = 1) -> None:
    """Run a function in parallel over a list of arguments."""
    logger = logging.getLogger(__name__)

    queue = Queue()

    arguments_chunked = generate_chunks(arguments, num_workers)
    for args in arguments_chunked:
        processes = [
            Process(
                target=target,
                args=(function, arg, queue),
                name=get_error_message_prefix(arg),
            )
            for arg in args
        ]

        for process in processes:
            process.start()

        for process in processes:
            process.join()
            if process.exitcode == 1:
                logger.error(queue.get())
                break



from __future__ import annotations

from collections import defaultdict
from pathlib import Path

import numpy as np
from sklearn.neighbors import LocalOutlierFactor

from DAJIN2.utils import io

"""
Sequence errors (such as large deletions) present in the control significantly impair the accuracy of mutation_loci. Therefore, only reads similar to alleles after clustering are desired for analysis.
Additionally, the mutation regions in the alleles after clustering should be identified and masked.
"""


def onehot_by_mutations(midsv_sample: list[dict]) -> dict[str, np.ndarray]:
    mut_onehot = defaultdict(list)
    for c in midsv_sample:
        cssplits = c["CSSPLIT"].split(",")
        for mut in {"+", "-", "*"}:
            onehot = [1 if cs.startswith(mut) else 0 for cs in cssplits]
            mut_onehot[mut].append(onehot)
    return {mut: np.array(value) for mut, value in mut_onehot.items()}


def calculate_percentage(
    mut_onehot_sample: dict[str, np.ndarray], coverage_match: np.ndarray[int]
) -> dict[str, np.ndarray]:
    mut_percentage = {}
    for mut, onehot in mut_onehot_sample.items():
        x = np.sum(onehot, axis=0) / coverage_match
        mut_percentage[mut] = np.where(np.isnan(x), 0, x)
    return mut_percentage


def get_values_to_mask(mut_percentage_sample: dict[str, np.ndarray], threshold=0.5) -> dict[str, np.ndarray[float]]:
    mask = {}
    for mut, percentage in mut_percentage_sample.items():
        mask[mut] = np.where(percentage > threshold, 0, percentage)
    return mask


def apply_mask(mut_onehot: dict[str, np.ndarray], mask_sample: dict[str, np.ndarray[float]]):
    mut_onehot_masked = {}
    for mut, onehot in mut_onehot.items():
        mut_onehot_masked[mut] = onehot * mask_sample[mut]
    return mut_onehot_masked


def identify_normal_reads(
    mut_onehot_sample_masked: dict[str, np.ndarray], mut_onehot_control_masked: dict[str, np.ndarray]
) -> list[bool]:
    mutation_comparisons = {}
    for mut in {"+", "-", "*"}:
        values_sample = mut_onehot_sample_masked[mut]
        values_control = mut_onehot_control_masked[mut]

        values_sum_sample = values_sample.sum(axis=1).reshape(-1, 1)
        values_sum_control = values_control.sum(axis=1).reshape(-1, 1)

        clf = LocalOutlierFactor(novelty=True, n_neighbors=len(values_sum_sample)).fit(values_sum_sample)
        labels = clf.predict(values_sum_control)
        mutation_comparisons[mut] = np.where(labels == 1, True, False)

    return (mutation_comparisons["+"] * mutation_comparisons["-"] * mutation_comparisons["*"]).tolist()


###########################################################
# main
###########################################################


def filter_control(ARGS, path_consensus_control: Path, path_consensus_sample: Path, allele: str) -> list[bool]:
    """
    find similar control reads compared to sample reads
    """
    cssplits = (m["CSSPLIT"].split(",") for m in io.read_jsonl(path_consensus_sample))
    coverage_match = np.array([sum(1 for cs in cssplit if cs.startswith("=")) for cssplit in zip(*cssplits)])
    mut_onehot_sample = onehot_by_mutations(io.read_jsonl(path_consensus_sample))

    path_mut_onehot_control = Path(
        ARGS.tempdir, ARGS.control_name, "consensus", allele, f"{ARGS.sample_name}_onehot.pickle"
    )
    if path_mut_onehot_control.exists():
        mut_onehot_control = io.load_pickle(path_mut_onehot_control)
    else:
        mut_onehot_control = onehot_by_mutations(io.read_jsonl(path_consensus_control))
        io.save_pickle(mut_onehot_control, path_mut_onehot_control)

    mut_percentage_sample = calculate_percentage(mut_onehot_sample, coverage_match)
    values_mask = get_values_to_mask(mut_percentage_sample)

    mut_onehot_sample_masked = apply_mask(mut_onehot_sample, values_mask)
    mut_onehot_control_masked = apply_mask(mut_onehot_control, values_mask)

    return identify_normal_reads(mut_onehot_sample_masked, mut_onehot_control_masked)


def cache_selected_control_by_similarity(
    ARGS, path_consensus_control: Path, path_consensus_sample: Path, allele: str
) -> None:
    normal_reads_flags = filter_control(ARGS, path_consensus_control, path_consensus_sample, allele)
    midsv_control = io.read_jsonl(path_consensus_control)
    midsv_filtered = (m for m, flag in zip(midsv_control, normal_reads_flags) if flag is True)

    io.write_jsonl(midsv_filtered, Path(path_consensus_sample.parent, "control.jsonl"))



from __future__ import annotations

from itertools import groupby
from pathlib import Path
from typing import Generator

import numpy as np
from rapidfuzz import process
from rapidfuzz.distance import DamerauLevenshtein
from sklearn.cluster import MeanShift

from DAJIN2.core.consensus.consensus import call_percentage
from DAJIN2.core.preprocess.sv_handler import add_unique_allele_keys, extract_unique_sv, save_cstag, save_fasta
from DAJIN2.utils import config, io
from DAJIN2.utils.cssplits_handler import convert_cssplit_to_dna, convert_cssplits_to_cstag, revcomp_cssplits

config.set_warnings_ignore()

import cstag

###########################################################
# Detect insertion sequences
###########################################################


def extract_inversions(midsv_sample: Generator[list[str]]) -> list[dict[int, int, str, str]]:
    """Extract sequence of inversions and their start and end indices."""
    inversions = []
    for m_sample in midsv_sample:
        cssplits = m_sample["CSSPLIT"].split(",")
        min_idx = float("inf")
        max_idx = -1
        cs_inversion = []
        for i, cs in enumerate(cssplits):
            if cs[-1].islower():
                min_idx = min(min_idx, i)
                max_idx = max(max_idx, i)
                cs_inversion.append(cs)
        if cs_inversion:
            inversions.append(
                {
                    "start": min_idx,
                    "end": max_idx,
                    "seq": convert_cssplit_to_dna(",".join(cs_inversion)),
                    "CSSPLIT": m_sample["CSSPLIT"],
                }
            )

    return inversions


def _convert_sequences_to_distances(sequences: list[str]) -> list[float]:
    """Convert sequences to distances using Damerau-Levenshtein distance."""
    query = sequences[0]
    _, distances, _ = zip(*process.extract_iter(query, sequences, scorer=DamerauLevenshtein.normalized_distance))
    return distances


def _convert_to_scores(start_end: Generator[int, int], distances: list[float]) -> list[int, int, float]:
    scores = []
    for (start, end), dist in zip(start_end, distances):
        scores.append([start, end, dist])
    return scores


def clustering_inversions(inversions: list[dict[int, int, str, str]], coverage: int) -> list[int]:
    distances = _convert_sequences_to_distances([d["seq"] for d in inversions])
    scores = _convert_to_scores(((d["start"], d["end"]) for d in inversions), distances)

    X = np.array(scores)
    min_size = max(5, int(coverage * 0.5 / 100))

    return MeanShift(min_bin_freq=min_size, bin_seeding=True).fit_predict(X).tolist()


###########################################################
# Call consensus
###########################################################


def call_consensus_of_inversion(
    inversions: list[dict[int, int, str, str]], labels: list[int], mutation_loci: list[set[str]], sequence: str
) -> dict[int, dict[str, str]]:
    """Generate consensus cssplits."""

    consensus_of_inversions = {}

    # Append labels
    for i, d in enumerate(inversions):
        d["label"] = labels[i]

    inversions.sort(key=lambda x: x["label"])
    for label, group in groupby(inversions, key=lambda x: x["label"]):
        group = list(group)[:1000]  # Downsample to 1000 sequences
        cssplits = [cs["CSSPLIT"].split(",") for cs in group]
        cons_percentage = call_percentage(cssplits, mutation_loci, sequence)
        cons_cssplits = [max(cons_per, key=cons_per.get) for cons_per in cons_percentage]

        # Detect inversion regions
        min_idx = float("inf")
        max_idx = -1
        for i, cs in enumerate(cons_cssplits):
            if cs[-1].islower():
                min_idx = min(min_idx, i)
                max_idx = max(max_idx, i)

        cons_inversion = revcomp_cssplits(cons_cssplits[min_idx : max_idx + 1])
        cons_cssplits_inversion_corrected = cons_cssplits[:min_idx] + cons_inversion + cons_cssplits[max_idx + 1 :]

        cons_cstag = convert_cssplits_to_cstag(cons_cssplits_inversion_corrected)
        cons_sequence = []
        cons_sequence = cstag.to_sequence(cons_cstag)
        cons_cssplits_inversion_corrected = ",".join(cons_cssplits_inversion_corrected)

        consensus_of_inversions[label] = {
            "CSTAG": cons_cstag,
            "SEQ": cons_sequence,
            "CSSPLIT": cons_cssplits_inversion_corrected,
        }

    return consensus_of_inversions


###########################################################
# main
###########################################################


def detect_inversions(TEMPDIR, SAMPLE_NAME, CONTROL_NAME, FASTA_ALLELES) -> None:
    path_sample = Path(TEMPDIR, SAMPLE_NAME, "midsv", "control", f"{SAMPLE_NAME}.jsonl")
    # path_control = Path(TEMPDIR, CONTROL_NAME, "midsv", "control", f"{CONTROL_NAME}.jsonl")
    mutation_loci = io.load_pickle(Path(TEMPDIR, SAMPLE_NAME, "mutation_loci", "control", "mutation_loci.pickle"))
    sequence = FASTA_ALLELES["control"]

    inversions = extract_inversions(io.read_jsonl(path_sample))
    if not inversions:  # If there is no insertion, return None
        return None

    coverage: int = io.count_newlines(path_sample)

    labels = clustering_inversions(inversions, coverage)

    consensus_of_inversions = call_consensus_of_inversion(inversions, labels, mutation_loci, sequence)

    seq_inversions = {key: d["SEQ"] for key, d in consensus_of_inversions.items()}
    consensus_of_unique_inversions = {
        key: consensus_of_inversions[key] for key in extract_unique_sv(seq_inversions, FASTA_ALLELES)
    }
    if consensus_of_unique_inversions == {}:
        return None

    consensus_of_unique_inversions = add_unique_allele_keys(consensus_of_unique_inversions, FASTA_ALLELES, "inversion")

    save_fasta(TEMPDIR, SAMPLE_NAME, {key: d["SEQ"] for key, d in consensus_of_unique_inversions.items()})
    save_cstag(TEMPDIR, SAMPLE_NAME, {key: d["CSTAG"] for key, d in consensus_of_unique_inversions.items()})

    # ToDo: Use CSSPLIT to reflect inversion in HTML
    # ToDo: Subtract from Sample when Control also has inversion



from __future__ import annotations

import hashlib
import os
import ssl
import xml.etree.ElementTree as ET
from pathlib import Path
from urllib.error import URLError
from urllib.request import urlopen

import mappy
import pysam

########################################################################
# To accommodate cases where a user might input negative values or
# excessively large values, update the number of threads
# from "1" to "max cpu threads - 1".
########################################################################


def update_threads(threads: int) -> int:
    available_threads = os.cpu_count() - 1
    threads_updated = max(1, min(threads, available_threads))
    return threads_updated


########################################################################
# Check if the sample is in the proper format.
########################################################################


def validate_file_existence(path_file: str) -> None:
    if not Path(path_file).exists():
        raise FileNotFoundError(f"{path_file} is not found")


def return_file_extension(path_file: Path) -> str:
    extensions = [".fastq", ".fastq.gz", ".fq", ".fq.gz", ".fasta", ".fasta.gz", ".fa", ".fa.gz", ".bam"]
    file_suffixes = "".join(path_file.suffixes)  # Combine all suffixes into a single string

    for ext in extensions:
        if file_suffixes == ext:
            return ext

    raise ValueError(
        f"{path_file} requires extensions either .fastq, .fastq.gz, .fq, .fq.gz, .fasta, .fasta.gz, .fa, .fa.gz, or .bam"
    )


# Varidate if the file is in the proper format viewing top 100 lines
def validate_fastq_content(path_fastq: str) -> None:
    try:
        headers, seqs, quals = zip(*[(n, s, q) for i, (n, s, q) in enumerate(mappy.fastx_read(path_fastq)) if i < 100])
        # Remove empty elements
        headers = [header for header in headers if header]
        seqs = [seq for seq in seqs if seq]
        quals = [qual for qual in quals if qual]

        if not (len(headers) == len(seqs) == len(quals) > 0):
            raise ValueError

    except ValueError:
        raise ValueError(f"{path_fastq} is not a proper FASTQ format")


def validate_fasta_content(path_fasta: str, allele_file: bool = False) -> None:
    try:
        headers, seqs = zip(*[(n, s) for n, s, _ in mappy.fastx_read(path_fasta)])
        # Remove empty elements
        headers = [header for header in headers if header]
        seqs = [seq for seq in seqs if seq]

        if len(headers) != len(seqs) or not headers:
            raise ValueError

    except ValueError:
        raise ValueError(f"{path_fasta} is not a proper FASTA format")

    if len(headers) != len(set(headers)):
        raise ValueError(f"{path_fasta} must include unique identifiers")

    if len(seqs) != len(set(seqs)):
        raise ValueError(f"{path_fasta} must include unique DNA sequences")

    if allele_file and "control" not in headers:
        raise ValueError(f"One of the headers in the {path_fasta} must be '>control'")


def validate_bam_content(path_bam: str) -> None:
    try:
        _ = pysam.AlignmentFile(path_bam, "rb")
    except ValueError:
        raise ValueError(f"{path_bam} is not a proper BAM format")


def validate_files(SAMPLE: str, CONTROL: str, ALLELE: str) -> None:
    for path_directory in [CONTROL, SAMPLE]:
        extentions = {return_file_extension(path_fastx) for path_fastx in Path(path_directory).iterdir()}
        if len(extentions) == 1:
            extention = next(iter(extentions))
        else:
            raise ValueError(
                f"{path_directory} contains multiple extensions. Please check if there are any incorrect files."
            )

        fasta_extensions = {".fasta", ".fa", ".fasta.gz", ".fa.gz"}
        fastq_extentions = {".fastq", ".fq", ".fastq.gz", ".fq.gz"}
        if extention in fasta_extensions:
            _ = [validate_fasta_content(str(path_fastx)) for path_fastx in Path(path_directory).iterdir()]
        elif extention in fastq_extentions:
            _ = [validate_fastq_content(str(path_fastx)) for path_fastx in Path(path_directory).iterdir()]
        else:
            _ = [validate_bam_content(str(path_bam)) for path_bam in Path(path_directory).iterdir()]

    validate_file_existence(ALLELE)
    validate_fasta_content(ALLELE, allele_file=True)


########################################################################
# Check Cache
########################################################################


def exists_cached_control(control: str, tempdir: Path) -> bool:
    path_cache_hash = Path(tempdir, "cache", "control_hash.txt")
    if path_cache_hash.exists():
        current_hash = hashlib.sha256(Path(control).read_bytes()).hexdigest()
        cashed_hash = path_cache_hash.read_text()
        if current_hash == cashed_hash:
            return True
    return False


def exists_cached_genome(genome: str, tempdir: Path, exists_cache_control: bool) -> bool:
    path_cache_genome = Path(tempdir, "cache", "genome_symbol.txt")
    if genome and exists_cache_control and path_cache_genome.exists():
        cashed_genome = path_cache_genome.read_text()
        if genome == cashed_genome:
            return True
    return False


########################################################################
# Check genome and UCSC server
########################################################################


def fetch_html_without_verification(url: str) -> str:
    context = ssl._create_unverified_context()  # Create an SSL context that temporarily disables verification
    with urlopen(url, context=context, timeout=10) as response:
        return response.read().decode("utf-8")


def format_url(key: str, url: str) -> str:
    return url + "/hg38" if key == "goldenpath" else url


def get_first_available_url(key: str, urls: list[str]) -> str | None:
    search_keys = {"blat": "BLAT Search Genome", "das": "GRCh38/hg38", "goldenpath": "bigZips"}
    return next(
        (url for url in urls if search_keys[key] in fetch_html_without_verification(format_url(key, url))), None
    )


def fetch_xml_without_verification(url: str) -> bytes:
    """Fetch XML data from a given URL."""
    context = ssl._create_unverified_context()  # Create an SSL context that temporarily disables verification
    with urlopen(url, context=context, timeout=10) as response:
        return response.read()


def extract_genome_ids_from_xml(xml_data: bytes) -> set[str]:
    """Extract genome IDs from XML data."""
    root = ET.fromstring(xml_data)
    return {cc.attrib["id"] for child in root for cc in child if cc.tag == "SOURCE"}


def get_genome_ids_in_ucsc(url_das: str) -> set[str]:
    """Get available genome IDs in UCSC."""
    xml_data = fetch_xml_without_verification(url_das)
    return extract_genome_ids_from_xml(xml_data)


def is_genome_id_available_in_ucsc(genome: str, url_das: str) -> bool:
    genome_ids = get_genome_ids_in_ucsc(url_das)
    return genome in genome_ids


def validate_genome_and_fetch_urls(genome: str) -> dict[str, str]:
    server_lists = {
        "blat": [
            "https://genome.ucsc.edu/cgi-bin/hgBlat",
            "https://genome-asia.ucsc.edu/cgi-bin/hgBlat",
            "https://genome-euro.ucsc.edu/cgi-bin/hgBlat",
        ],
        "das": [
            "https://genome.ucsc.edu/cgi-bin/das/dsn/",
            "https://genome-asia.ucsc.edu/cgi-bin/das/dsn/",
            "https://genome-euro.ucsc.edu/cgi-bin/das/dsn",
        ],
        "goldenpath": [
            "https://hgdownload.cse.ucsc.edu/goldenPath",
            "https://hgdownload.soe.ucsc.edu/goldenPath",
        ],
    }

    available_servers = {key: get_first_available_url(key, urls) for key, urls in server_lists.items()}

    error_messages = {
        "blat": "All UCSC blat servers are currently down. Please wait for a while and try again.",
        "das": "All UCSC DAS servers are currently down. Please wait for a while and try again.",
        "goldenpath": "All UCSC GoldenPath servers are currently down. Please wait for a while and try again.",
    }

    for key, message in error_messages.items():
        if available_servers[key] is None:
            raise URLError(message)

    if not is_genome_id_available_in_ucsc(genome, available_servers["das"]):
        raise ValueError(f"{genome} is not listed. Available genomes are in {available_servers['das']}")

    del available_servers["das"]

    return available_servers



from __future__ import annotations

from pathlib import Path
from typing import Generator

from DAJIN2.utils import io


def compress_insertion(cssplit: list[str], index: int, compress_ins: bool) -> str:
    """Process insertion mutations based on the compress_ins flag.
    compress_ins: True -> +I{inserted sequence}
    compress_ins: False -> +{number of |}{inserted sequence}

    Note: Clustering is based on the number of nucleotides inserted, but this creates too fine a cluster, so the number of nucleotides is ignored.Ideally, clustering should be possible without ignoring the number of bases.
    """
    if compress_ins:
        for j in range(index - 1, index + 2):
            if cssplit[j].startswith("+") and not cssplit[j].startswith("+I"):
                cssplit[j] = f"+I{cssplit[j].split('|')[-1]}"
    else:
        for j in range(index - 1, index + 2):
            if cssplit[j].startswith("+") and not cssplit[j][1].isdigit():
                cssplit[j] = f"+{cssplit[j].count('|')}{cssplit[j].split('|')[-1]}"
    return cssplit


def annotate_cs_tag(cssplit: str, mutation: set[str]) -> str:
    cs = cssplit[0]  # +, - , *, =, or N
    if cs in mutation:
        return cssplit
    else:
        return "@"


def generate_mutation_kmers(
    path_sample: Path | str, mutation_loci: list[set[str]], compress_ins: bool = True
) -> Generator[list[str]]:
    midsv_sample = io.read_jsonl(path_sample)
    for cssplit in (cs["CSSPLIT"].split(",") for cs in midsv_sample):
        mutation_kmers = ["@,@,@"]
        for i in range(1, len(cssplit) - 1):
            if mutation_loci[i] == set():
                mutation_kmers.append("@,@,@")
                continue
            cs_current = cssplit[i][0]  # +, - , *, =, N
            if cs_current == "+":
                cssplit = compress_insertion(cssplit, i, compress_ins)
            if cs_current in mutation_loci[i]:
                cs_prev = annotate_cs_tag(cssplit[i - 1], mutation_loci[i - 1])
                cs_next = annotate_cs_tag(cssplit[i + 1], mutation_loci[i + 1])
                kmer = ",".join([cs_prev, cssplit[i], cs_next])
                mutation_kmers.append(kmer)
            else:
                mutation_kmers.append("@,@,@")
        mutation_kmers.append("@,@,@")
        yield mutation_kmers



from DAJIN2.core.consensus.clust_formatter import downsample_by_label, remove_minor_alleles
from DAJIN2.core.consensus.consensus import call_consensus
from DAJIN2.core.consensus.mutation_extractor import cache_mutation_loci
from DAJIN2.core.consensus.name_handler import add_key_by_allele_name, call_allele_name, update_key_by_allele_name



from __future__ import annotations

from collections import Counter
from itertools import chain
from pathlib import Path

import numpy as np
from scipy.sparse import csr_matrix, spmatrix
from sklearn import metrics
from sklearn.cluster import BisectingKMeans

from DAJIN2.core.clustering.constrained_kmeans import ConstrainedKMeans
from DAJIN2.core.clustering.score_handler import subset_scores
from DAJIN2.core.clustering.strand_bias_handler import remove_biased_clusters
from DAJIN2.utils import config, io

config.set_warnings_ignore()


###############################################################################
# Clustering
###############################################################################


def count_number_of_clusters(labels_control: list[int], coverage_control: int) -> int:
    """If there is less than 1% lead within a cluster, they are considered clustering errors and are not counted"""
    return sum(1 for control_reads in Counter(labels_control).values() if control_reads / coverage_control > 0.01)


def optimize_labels(X: spmatrix, coverage_sample: int, coverage_control: int, min_cluster_size: int) -> list[int]:
    labels_previous = [1 for _ in range(coverage_sample)]
    positive_silhouette_previous = 1

    for i in range(2, coverage_sample):
        np.random.seed(seed=1)
        labels_all = (
            ConstrainedKMeans(n_clusters=i, min_cluster_size=min_cluster_size, random_state=1).fit_predict(X).tolist()
        )

        labels_sample = labels_all[:coverage_sample]
        labels_control = labels_all[coverage_sample:]

        min_cluster_size_ = min(Counter(labels_sample).values())

        num_labels_control = count_number_of_clusters(labels_control, coverage_control)

        silhouette_scores = metrics.silhouette_samples(X, labels_all, metric="euclidean")
        positive_silhouette_current = sum(silhouette_scores > 0.25)
        silhouette_ratio = positive_silhouette_current / positive_silhouette_previous

        # print(
        #     i,
        #     min_cluster_size,
        #     Counter(labels_control),
        #     Counter(labels_sample),
        #     silhouette_ratio,
        #     positive_silhouette_current,
        #     positive_silhouette_previous,
        # )  # ! DEBUG

        if min_cluster_size_ < min_cluster_size - 1 or num_labels_control >= 2 or silhouette_ratio < 0.95:
            break

        labels_previous = labels_sample
        positive_silhouette_previous = max(positive_silhouette_previous, positive_silhouette_current)
    return labels_previous


def get_label_most_common(labels: list[int]) -> int:
    return Counter(labels).most_common()[0][0]


def return_labels(
    path_score_sample: Path,
    path_score_control: Path,
    path_sample: Path,
    strand_bias_in_control: bool,
    min_cluster_size: int,
) -> list[int]:
    np.random.seed(seed=1)

    score_control = list(io.read_jsonl(path_score_control))
    X_control = csr_matrix(score_control)

    """Subset to 1000 reads of controls in the most common cluster to remove outliers and reduce computation time"""
    labels_control = BisectingKMeans(n_clusters=2, random_state=1).fit_predict(X_control)
    label_most_common = get_label_most_common(labels_control)
    scores_control_subset = subset_scores(labels_control, io.read_jsonl(path_score_control), label_most_common, 1000)

    scores_sample = list(io.read_jsonl(path_score_sample))
    X = np.array(list(chain(scores_sample, scores_control_subset)))

    coverage_sample = io.count_newlines(path_score_sample)
    coverage_control = len(scores_control_subset)
    labels = optimize_labels(X, coverage_sample, coverage_control, min_cluster_size)

    """Re-allocate clusters with strand bias to clusters without strand bias"""
    if strand_bias_in_control is False:
        labels = remove_biased_clusters(path_sample, path_score_sample, labels)

    return labels



from __future__ import annotations

from itertools import groupby
from pathlib import Path

from DAJIN2.utils.cssplits_handler import reallocate_insertion_within_deletion, revcomp_cssplits

###########################################################
# group by mutation
###########################################################


def annotate_inversion(cssplits: list[str]) -> list[str]:
    return ["@" + cs if cs[-1].islower() else cs for cs in cssplits]


def group_by_mutation(cssplits: list[str]) -> list[list[str]]:
    return [list(group) for _, group in groupby(cssplits, key=lambda x: x[0])]


###########################################################
# Report mutations
###########################################################


def _handle_match(group, genome, start, end, header, chromosome) -> tuple[list[list[str]], int, int]:
    end += len(group)
    start = end
    return [None], start, end


def _handle_substitution(group, genome, start, end, header, chromosome) -> tuple[list[list[str]], int, int]:
    result = []
    for g in group:
        ref = g[1]
        mut = g[2]
        result.append([header, genome, chromosome, start, end, f"substitution: {ref}>{mut}"])
        end += 1
        start = end
    return result, start, end


def _handle_deletion(group, genome, start, end, header, chromosome) -> tuple[list[list[str]], int, int]:
    end += len(group) - 1
    size = len(group)
    seq = "".join([g[-1] for g in group])
    result = [header, genome, chromosome, start, end, f"{size}bp deletion: {seq}"]
    end += 1
    start = end
    return [result], start, end


def _handle_insertion(group, genome, start, end, header, chromosome) -> tuple[list[list[str]], int, int]:
    group = group[0]
    size = group.count("|")
    seq_insertion = "".join([g[-1] for g in group.split("|")[:-1]])
    seq_last = group.split("|")[-1]
    result = []
    result.append([header, genome, chromosome, start, end, f"{size}bp insertion: {seq_insertion}"])
    if seq_last.startswith("="):
        pass
    elif seq_last.startswith("-"):
        result.append([header, genome, chromosome, start, end, f"1bp deletion: {seq_last[-1]}"])
    end += 1
    start = end
    return result, start, end


def _handle_inversion(group, genome, start, end, header, chromosome) -> tuple[list[list[str]], int, int]:
    end += len(group) - 1
    size = len(group)
    seq = "".join([g[-1].upper() for g in group])
    result = [header, genome, chromosome, start, end, f"{size}bp inversion: {seq}"]
    end += 1
    start = end
    return [result], start, end


def _handle_unknown(group, genome, start, end, header, chromosome) -> tuple[list[list[str]], int, int]:
    end += len(group) - 1
    size = len(group)
    result = [header, genome, chromosome, start, end, f"{size}bp unknown bases"]
    end += 1
    start = end
    return [result], start, end


def report_mutations(cssplits_grouped: list[list[str]], GENOME_COORDINATES, header) -> list[list[str]]:
    genome = GENOME_COORDINATES["genome"]
    chromosome = GENOME_COORDINATES["chrom"]
    start = end = GENOME_COORDINATES["start"]
    handlers = {
        "=": _handle_match,
        "*": _handle_substitution,
        "-": _handle_deletion,
        "+": _handle_insertion,
        "@": _handle_inversion,
        "N": _handle_unknown,
    }
    results = []
    for group in cssplits_grouped:
        for prefix, handler in handlers.items():
            if group[0].startswith(prefix):
                result, start, end = handler(group, genome, start, end, header, chromosome)
                if prefix == "=":
                    continue
                results.extend(result)
    return [list(map(str, r)) for r in results]


###########################################################
# main
###########################################################


def export_to_csv(TEMPDIR: Path, SAMPLE_NAME: str, GENOME_COORDINATES: dict, cons_percentage: dict[str, list]) -> None:
    results = [["Allele ID", "Genome", "Chromosome", "Start", "End", "Mutation"]]
    for header, cons in cons_percentage.items():
        cssplits = [max(c, key=c.get) for c in cons]
        if GENOME_COORDINATES.get("strand") == "-":
            cssplits = revcomp_cssplits(cssplits)
        cssplits = reallocate_insertion_within_deletion(cssplits, bin_size=500, percentage=50)
        cssplits_inversion = annotate_inversion(cssplits)
        cssplits_grouped = group_by_mutation(cssplits_inversion)
        result = report_mutations(cssplits_grouped, GENOME_COORDINATES, header)
        results.extend(result)

    results_csv = "\n".join([",".join(map(str, r)) for r in results]) + "\n"

    path_output = Path(TEMPDIR, "report", "MUTATION_INFO", f"{SAMPLE_NAME}.csv")
    path_output.write_text(results_csv)
